/*
 * Spectator - a desktop app which communicates with router Turris
 * Copyright (C) 2015  CZ.NIC, z. s. p. o. <http://www.nic.cz>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "application.h"
#include <QtCore>
#include "logging.h"

SpectatorApplication::SpectatorApplication(int &argc, char **argv) :
	QApplication(argc, argv),
	forceSystrayMode(false),
	noSystrayMode(false)
{
	// Parse command line
	QCommandLineParser parser;
	parser.setApplicationDescription("Spectator");
	parser.addHelpOption();
	parser.addVersionOption();

	// -s --force-systray
	QCommandLineOption forceSystray = QCommandLineOption(
		QStringList() << "s" << "force-systray",
		"force to use systay mode"
	);
	parser.addOption(forceSystray);

	// -n --no-systray
	QCommandLineOption noSystray = QCommandLineOption(
		QStringList() << "n" << "no-systray",
		"force to use systay mode"
	);
	parser.addOption(noSystray);

	parser.process(*this);
	this->forceSystrayMode = parser.isSet(forceSystray);
	this->noSystrayMode = parser.isSet(noSystray);

	if (this->forceSystrayMode)
		log_settings.debug() << "Forcing to use systray mode";
	if (this->noSystrayMode)
		log_settings.debug() << "Forcing not to use systray mode";

}

int SpectatorApplication::exec()
{
	if(this->forceSystrayMode && this->noSystrayMode) {
		QTextStream out(stdout);
		out << "Can't use --no-systray and --force-systray options together!" << endl;
		return 1;
	}

	this->createSettings();
	this->createTrayIcon();

	// Main window should intialized last
	this->createMainWindow();

	// toggle main window on click
	QObject::connect(this->trayIcon, SIGNAL(activated(QSystemTrayIcon::ActivationReason)), this->mainWindow, SLOT(toggleWindow(QSystemTrayIcon::ActivationReason)));

	this->updateTranslations();
	return QApplication::exec();
}

void SpectatorApplication::createTrayIcon()
{
	// Create icon
	QIcon icon(":/icons/img/trayicon-gray.png");

	// Create systemtray with the icon
	this->trayIcon = new QSystemTrayIcon(icon, this);

	// tooltip
	this->trayIcon->setToolTip("Spectator");

	// prepare refreshing timer
	this->trayIconTimer = new QTimer(this);
	this->trayIconTimer->setSingleShot(false);
	this->trayIconTimer->setInterval(5000);
	QObject::connect(this->trayIconTimer, SIGNAL(timeout()), this, SLOT(refreshTrayIcon()));
	this->trayIconTimer->start();
}

void SpectatorApplication::createMainWindow()
{
	this->mainWindow = new MainWindow(NULL, this->trayIcon);
	// TODO share menu  in a nicer way
	this->trayIcon->setContextMenu(this->mainWindow->menu);
	this->mainWindow->useSettings(this->settings);
	if (this->mainWindow->isInSystrayMode()) {
		this->mainWindow->showRunningInSystrayPopup();
	}
	QObject::connect(this->mainWindow, SIGNAL(translationNeeded()), this, SLOT(updateTranslations()));
}

void SpectatorApplication::createSettings()
{
	this->settings = new SpectatorSettings();
	settings->load();
	settings->forceSystrayMode = this->forceSystrayMode;
	settings->noSystrayMode = this->noSystrayMode;
}

void SpectatorApplication::updateTranslations()
{
	// Remove translators if needed
	if (this->qtTranslator) {
		delete this->qtTranslator;
	}
	if (this->spectatorTranslator) {
		delete this->spectatorTranslator;
	}

	// Manage translations
	this->qtTranslator = new QTranslator(this);
	log_settings.debug() << "Installing translator:" << "qtbase_" + settings->language;
	log_settings.debug() << "Translator installed:" << this->qtTranslator->load("qtbase_" + settings->language, ":/translations/lang");
	qApp->installTranslator(this->qtTranslator);

	this->spectatorTranslator = new QTranslator(this);
	log_settings.debug() << "Installing translator:" << "spectator_" + settings->language;
	log_settings.debug() << "Translator installed:" << this->spectatorTranslator->load("spectator_" + settings->language, ":/translations/lang");
	qApp->installTranslator(this->spectatorTranslator);

	QString locale = settings->language == "cs" ? "cs_CZ": "en_US";
	log_settings.debug() << "Setting locale: " << locale;
	QLocale::setDefault(QLocale(locale));
	this->mainWindow->setLocale(QLocale(locale));
}

void SpectatorApplication::refreshTrayIcon()
{
	// try to show systemtray icon
	// note that some environments might pass the testSystrayMode() after a few seconds
	// the start of the desktop environment
	if (this->settings->systrayMode && this->mainWindow->testSystrayMode()) {
		this->mainWindow->setSystrayMode(true);
	}

}

SpectatorApplication::~SpectatorApplication()
{
	if (this->spectatorTranslator)
		delete this->spectatorTranslator;
	if (this->qtTranslator)
		delete this->qtTranslator;
	if (this->trayIconTimer)
		delete this->trayIconTimer;
	if (this->mainWindow)
		delete this->mainWindow;
	if (this->trayIcon)
		delete this->trayIcon;
	if (this->settings)
		delete this->settings;
}
