/*
 * Spectator - a desktop app which communicates with router Turris using NETCONF protocol
 * Copyright (C) 2015  CZ.NIC, z. s. p. o. <http://www.nic.cz>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <QtGlobal>
#include <QByteArray>
#include <QDateTime>
#include <QLabel>

#include "chat.h"

using Chats::Base;
using Chats::Initial;
using Chats::GetNotifications;
using Chats::DeleteNotification;
using Chats::RebootRouter;
using Chats::GetNethist;
using Chats::SetPassword;
using Chats::StoreBackup;
using Chats::RestoreBackup;
using Chats::GetStats;
using Chats::GetSerial;
using Chats::GetUpdaterData;
using Chats::CheckUpdater;
using Chats::SetUpdater;
using Chats::GetNeighbours;

namespace ChatStatus {

	enum statuses {
		started,
		quitting
	};

}

static inline QDomElement buildXmlHeaderAndRpc(QDomDocument &xml, int rpcMessageId)
{
	// xml append header
	xml.appendChild(xml.createProcessingInstruction( "xml", "version=\"1.0\" encoding=\"UTF-8\""));
	// append root element
	QDomElement root = xml.createElementNS("urn:ietf:params:xml:ns:netconf:base:1.0", "rpc");
	root.setAttribute("message-id", rpcMessageId);
	xml.appendChild(root);
	return root;
}

Base::Base()
{
	qsrand(QDateTime::currentMSecsSinceEpoch());
	this->rpcMessageId = qrand();
	this->status = ChatStatus::started;
}

bool Base::parseXml(QDomDocument &xml, QString msg)
{
	if (!xml.setContent(msg, true)) {
		emit this->report(Communicator::RT_ERROR, tr("Failed to parse obtained data."));
		emit this->finished(false);
		return false;
	}
	return true;
}

bool Base::checkRpcMessageId(const QDomDocument &xml)
{
	QDomNodeList rpcReplies = xml.elementsByTagName("rpc-reply");
	if (rpcReplies.isEmpty()) {
		emit this->report(Communicator::RT_ERROR, tr("Failed to parse RPC reply message."));
		emit this->finished(false);
		return false;
	}
	QDomAttr attr = rpcReplies.at(0).attributes().namedItem("message-id").toAttr();
	if (attr.isNull() || attr.value() != QString::number(this->rpcMessageId)) {
		emit this->report(Communicator::RT_ERROR, tr("RPC message-id mismatch."));
		emit this->finished(false);
		return false;
	}
	return true;
}

bool Base::checkRpcMessageOk(const QDomDocument &xml, bool report)
{
	QDomNode rpcReply = xml.elementsByTagName("rpc-reply").at(0);
	if (rpcReply.firstChild().localName() != "ok") {
		if (report) {
			emit this->report(Communicator::RT_ERROR, tr("Incorrect RPC response."));
			emit this->finished(false);
		}
		return false;
	}
	return true;
}

bool Base::parseNetconfErrors(const QDomDocument &xml, QList<NetconfError> &outErrors)
{
	QDomNodeList errorElements = xml.elementsByTagName("rpc-error");
	if (errorElements.isEmpty()) {
		return false;
	}

	for (int i = 0; i < errorElements.count(); ++i) {
		QDomNode node = errorElements.at(i);
		NetconfError record;
		QDomElement element;

		element = node.firstChildElement("error-type");
		if (!element.isNull())
			record.type = element.text();

		element = node.firstChildElement("error-tag");
		if (!element.isNull())
			record.tag = element.text();

		element = node.firstChildElement("error-severity");
		if (!element.isNull())
			record.serverity = element.text();

		element = node.firstChildElement("error-app-tag");
		if (!element.isNull())
			record.appTag = element.text();

		element = node.firstChildElement("error-path");
		if (!element.isNull())
			record.path = element.text();

		element = node.firstChildElement("error-message");
		if (!element.isNull())
			record.message = element.text();

		element = node.firstChildElement("error-info");
		QDomNodeList errorInfoList = element.childNodes();
		for (int j = 0; j < errorInfoList.count(); ++j) {
			QDomNode subnode = errorInfoList.at(j);
			NetconfErrorInfoElement errorInfo;
			errorInfo.name = subnode.localName();
			errorInfo.content = subnode.toElement().text();
			record.info.append(errorInfo);
		}
		outErrors.append(record);
	}

	return true;
}

void Base::finish(bool res)
{
	this->status = ChatStatus::quitting;
	emit this->finished(res);
}

QList<Capability> Initial::requiredCapabilities = QList<Capability>();

void Initial::talk()
{
	// nothing to be done here (waiting for the response here)
}

QString Initial::capibilityMessage()
{
	QDomDocument xml;

	// xml append header
	xml.appendChild(xml.createProcessingInstruction( "xml", "version=\"1.0\" encoding=\"UTF-8\""));
	// append root element
	QDomElement root = xml.createElementNS("urn:ietf:params:xml:ns:netconf:base:1.0", "hello");
	xml.appendChild(root);

	// append capabilities list
	QDomElement capabilities = xml.createElement("capabilities");
	root.appendChild(capabilities);

	// append capability
	QDomElement capability = xml.createElement("capability");
	capability.appendChild(xml.createTextNode("urn:ietf:params:netconf:base:1.0"));
	capabilities.appendChild(capability);

	return xml.toString();
}

Capability::Capability(QString url)
{
	// parse using url
	QUrl capabilityUrl(url);
	this->base = capabilityUrl.scheme() + "://" + capabilityUrl.host() + capabilityUrl.path();
	QUrlQuery urlQuery(capabilityUrl);
	this->revision = urlQuery.hasQueryItem("revision") ? urlQuery.queryItemValue("revision") : QString();
}


bool Capability::covers(Capability &capability)
{
	// base should always match
	if (this->base != capability.base)
		return false;

	// NULL revision means any revision
	if (capability.revision.isNull())
		return true;

	// check the revsions
	if (this->revision != capability.revision)
		return false;

	return true;
}

void Initial::readMessage(QString msg)
{
	// Parse xml
	QDomDocument xml;
	if (!this->parseXml(xml, msg))
		return;

	// Parse capabilities
	QDomNodeList capabilities = xml.elementsByTagName("capability");
	if (!capabilities.isEmpty()) {
		QList<Capability> list;
		for (int i = 0; i < capabilities.count(); ++i) {
			list.append(Capability(capabilities.at(i).toElement().text()));
		}
		emit this->exportCapabilities(list);
	} else {
		emit this->exportCapabilities(QList<Capability>());
	}

	switch (status) {
		case ChatStatus::started:
			emit this->report(Communicator::RT_SUCCESS, tr("An initial message received."));
			emit this->writeMessage(Initial::capibilityMessage());
			emit this->communicationStatusUpdate(Communicator::CONNECTOR_READY);
			this->finish(true);
			break;

		case ChatStatus::quitting:
			this->finish(true);
			break;
	}
}


QList<Capability> GetNotifications::requiredCapabilities = QList<Capability>() << Capability(CAPABILITY_NOTIFICATION_BASE, QString());

QString GetNotifications::getNotificationsMessage(int rpcMessageId)
{
	QDomDocument xml;

	QDomElement root = buildXmlHeaderAndRpc(xml, rpcMessageId);

	// append get
	QDomElement get = xml.createElement("get");
	root.appendChild(get);

	// append filter
	QDomElement filter = xml.createElement("filter");
	filter.setAttribute("type", "subtree");
	get.appendChild(filter);

	QDomElement messages = xml.createElementNS(CAPABILITY_NOTIFICATION_BASE, "messages");
	filter.appendChild(messages);

	return xml.toString();
}

void GetNotifications::talk()
{
	emit this->writeMessage(GetNotifications::getNotificationsMessage(this->rpcMessageId));
}

bool compareNotifications(const ParsedNotification &n1, const ParsedNotification &n2)
{
	return n1.id < n2.id;
}

void GetNotifications::readMessage(QString msg)
{
	// Parse xml
	QDomDocument xml;
	if (!this->parseXml(xml, msg))
		return;

	// Check message id
	if (!this->checkRpcMessageId(xml))
		return;

	QDomNodeList messages = xml.elementsByTagName("message");
	if (!messages.isEmpty()) {
		QList<ParsedNotification> list;
		for (int i = 0; i < messages.count(); ++i) {
			if (messages.at(i).firstChildElement("displayed").isNull()) {  // filters removed
				ParsedNotification notification;
				notification.id = messages.at(i).firstChildElement("id").text();
				QDomElement bodyElement = messages.at(i).firstChildElement("body");
				QString noLangText;
				while (!bodyElement.isNull()) {
					if (bodyElement.attribute("lang") == this->lang) {
						notification.body = bodyElement.text();
						break;
					} else {
						noLangText = bodyElement.text();
					}
					bodyElement = bodyElement.nextSiblingElement("body");
				}

				if (notification.body.isNull())  // if the required lang is not present display at least some language
					notification.body = noLangText;

				notification.severity = messages.at(i).firstChildElement("severity").text();
				notification.time = QDateTime::fromTime_t(messages.at(i).firstChildElement("timestamp").text().toInt());
				list.append(notification);
			}
		}
		// sort notifications by id
		qSort(list.begin(), list.end(), compareNotifications);
		emit this->exportNotifications(list);
	} else {
		emit this->exportNotifications(QList<ParsedNotification>());
	}

	emit this->report(Communicator::RT_SUCCESS, tr("Notifications were downloaded."));
	this->finish(true);
}

QList<Capability> DeleteNotification::requiredCapabilities = QList<Capability>() << Capability(CAPABILITY_NOTIFICATION_BASE, QString());

QString DeleteNotification::deleteNotificationMessage(int rpcMessageId, QString messageId)
{
	QDomDocument xml;

	QDomElement root = buildXmlHeaderAndRpc(xml, rpcMessageId);

	// append display
	QDomElement display = xml.createElementNS(CAPABILITY_NOTIFICATION_BASE, "display");
	root.appendChild(display);

	// append filter
	QDomElement message_id = xml.createElement("message-id");
	message_id.appendChild(xml.createTextNode(messageId));
	display.appendChild(message_id);

	return xml.toString();
}

void DeleteNotification::talk()
{
	emit this->writeMessage(DeleteNotification::deleteNotificationMessage(this->rpcMessageId, this->messageId));
}

void DeleteNotification::readMessage(QString msg)
{
	// Parse xml
	QDomDocument xml;
	if (!this->parseXml(xml, msg))
		return;

	// Check message id
	if (!this->checkRpcMessageId(xml))
		return;

	// Check Ok
	bool res = this->checkRpcMessageOk(xml, false);
	if (!res) {
		// Look for rpc-error
		QList<NetconfError> errors;
		if (this->parseNetconfErrors(xml, errors))
			emit this->netconfErrorsOccured(errors);

		emit this->report(Communicator::RT_ERROR, tr("Incorrect RPC response."));
	} else {
		emit this->notificationDeleted(this->messageId);
	}

	this->finish(res);
}


QList<Capability> RebootRouter::requiredCapabilities = QList<Capability>() << Capability(CAPABILITY_MAINTAIN_BASE, QString());

QString RebootRouter::rebootRouterMessage(int rpcMessageId)
{
	QDomDocument xml;

	QDomElement root = buildXmlHeaderAndRpc(xml, rpcMessageId);

	// append display
	QDomElement display = xml.createElementNS(CAPABILITY_MAINTAIN_BASE, "reboot");
	root.appendChild(display);

	return xml.toString();
}

void RebootRouter::talk()
{
	emit this->writeMessage(RebootRouter::rebootRouterMessage(this->rpcMessageId));
}

void RebootRouter::readMessage(QString msg)
{
	// Parse xml
	QDomDocument xml;
	if (!this->parseXml(xml, msg))
		return;

	// Check message id
	if (!this->checkRpcMessageId(xml))
		return;

	// Check Ok
	bool res = this->checkRpcMessageOk(xml);
	if (!res) {
		// Look for rpc-error
		QList<NetconfError> errors;
		if (this->parseNetconfErrors(xml, errors))
			emit this->netconfErrorsOccured(errors);

		emit this->report(Communicator::RT_ERROR, tr("Incorrect RPC response."));
	} else {
		emit this->rebootingRouter();
	}

	this->finish(res);
}


QList<Capability> GetNethist::requiredCapabilities = QList<Capability>() << Capability(CAPABILITY_NETHIST_BASE, QString());

QString GetNethist::getNethistMessage(int rpcMessageId)
{
	QDomDocument xml;

	QDomElement root = buildXmlHeaderAndRpc(xml, rpcMessageId);

	// append get
	QDomElement get = xml.createElement("get");
	root.appendChild(get);

	// append filter
	QDomElement filter = xml.createElement("filter");
	filter.setAttribute("type", "subtree");
	get.appendChild(filter);

	QDomElement messages = xml.createElementNS(CAPABILITY_NETHIST_BASE, "nethist");
	filter.appendChild(messages);

	return xml.toString();
}

void GetNethist::talk()
{
	emit this->writeMessage(GetNethist::getNethistMessage(this->rpcMessageId));
}

bool compareNethist(const NethistRecord &n1, const NethistRecord &n2)
{
	return n1.time < n2.time;
}

void GetNethist::readMessage(QString msg)
{
	// Parse xml
	QDomDocument xml;
	if (!this->parseXml(xml, msg))
		return;

	// Check message id
	if (!this->checkRpcMessageId(xml))
		return;

	// Parse the nethist
	QDomNodeList snapshots = xml.elementsByTagName("snapshot");
	if (!snapshots.isEmpty()) {
		QList<NethistRecord> list;
		for (int i = 0; i < snapshots.count(); ++i) {
			NethistRecord record;
			if (!this->parseRecord(snapshots.at(i), record)) {
				emit this->report(Communicator::RT_ERROR, tr("Incorrect Nethist XML format."));
				this->finish(false);
				return;
			}
			list.append(record);
		}
		// sort records by time
		qSort(list.begin(), list.end(), compareNethist);

		emit this->exportNethist(list);
	} else {
		emit this->exportNethist(QList<NethistRecord>());
	}

	this->finish(true);
}

bool GetNethist::parseRecord(const QDomNode &node, NethistRecord &record)
{
	// parse time
	record.time = QDateTime::fromTime_t(node.firstChildElement("time").text().toInt());

	// parse network interfaces
	QDomNodeList interfaces = node.firstChildElement("network").childNodes();
	QList<NethistInterfaceRecord> interfaceRecords;
	for (int i = 0; i < interfaces.count(); ++i) {
		NethistInterfaceRecord interfaceRecord;
		interfaceRecord.name = interfaces.at(i).firstChildElement("name").text();
		interfaceRecord.rx = interfaces.at(i).firstChildElement("rx").text().toULongLong();
		interfaceRecord.tx = interfaces.at(i).firstChildElement("tx").text().toULongLong();
		interfaceRecords.append(interfaceRecord);
	}
	record.interfaces = interfaceRecords;

	// cpu
	record.cpuLoad = node.firstChildElement("cpu").firstChildElement("load").text().toDouble();

	// temperatore
	record.cpuTemperature = node.firstChildElement("temperature").firstChildElement("cpu").text().toInt();
	record.boardTemperature = node.firstChildElement("temperature").firstChildElement("board").text().toInt();

	// memory
	record.totalMemory = node.firstChildElement("memory").firstChildElement("memtotal").text().toInt();
	record.freeMemory = node.firstChildElement("memory").firstChildElement("memfree").text().toInt();
	record.buffersMemory = node.firstChildElement("memory").firstChildElement("buffers").text().toInt();
	record.cachedMemory = node.firstChildElement("memory").firstChildElement("cached").text().toInt();

	// disk
	record.usedDiskSpace = node.firstChildElement("rootfs").firstChildElement("used").text().toInt();
	record.availableDiskSpace = node.firstChildElement("rootfs").firstChildElement("available").text().toInt();

	return true;
}


QList<Capability> SetPassword::requiredCapabilities = QList<Capability>() << Capability(CAPABILITY_PASSWORD_BASE, QString());

SetPassword::~SetPassword()
{
	this->password.clear();
}

void SetPassword::readMessage(QString msg)
{
	// Parse xml
	QDomDocument xml;
	if (!this->parseXml(xml, msg))
		return;

	// Check message id
	if (!this->checkRpcMessageId(xml))
		return;

	// Check Ok
	bool res = this->checkRpcMessageOk(xml);
	if (!res) {
		// Look for rpc-error
		QList<NetconfError> errors;
		if (this->parseNetconfErrors(xml, errors))
			emit this->netconfErrorsOccured(errors);

		emit this->report(Communicator::RT_ERROR, tr("Incorrect RPC response."));
	}

	this->finish(res);
}

QString SetPassword::setPasswordMessage(int rpcMessageId, QString newPassword)
{
	QDomDocument xml;

	QDomElement root = buildXmlHeaderAndRpc(xml, rpcMessageId);

	// append set
	QDomElement set = xml.createElementNS(CAPABILITY_PASSWORD_BASE, "set");
	root.appendChild(set);

	// append user
	QDomElement user = xml.createElement("user");
	user.appendChild(xml.createTextNode("root"));
	set.appendChild(user);

	// append password
	QDomElement password = xml.createElement("password");
	password.appendChild(xml.createTextNode(newPassword));
	set.appendChild(password);

	return xml.toString();
}

void SetPassword::talk()
{
	emit this->writeMessage(SetPassword::setPasswordMessage(this->rpcMessageId, this->password.toHtmlEscaped()));
}


QList<Capability> StoreBackup::requiredCapabilities = QList<Capability>() << Capability(CAPABILITY_MAINTAIN_BASE, QString());

void StoreBackup::readMessage(QString msg)
{
	// Parse xml
	QDomDocument xml;
	if (!this->parseXml(xml, msg))
		return;

	// Check message id
	if (!this->checkRpcMessageId(xml))
		return;

	// Look for rpc-error
	QList<NetconfError> errors;
	if (this->parseNetconfErrors(xml, errors)) {
		emit this->netconfErrorsOccured(errors);
		this->finish(false);
		return;
	}

	bool res;
	QDomNodeList data = xml.elementsByTagName("data");
	if (data.length() == 1) {
		QByteArray base64_data(data.at(0).toElement().text().toStdString().c_str());
		emit this->storeData(QByteArray::fromBase64(base64_data));
		res = true;
	} else {
		res = false;
	}

	this->finish(res);
}

QString StoreBackup::storeBackupMessage(int rpcMessageId)
{
	QDomDocument xml;

	QDomElement root = buildXmlHeaderAndRpc(xml, rpcMessageId);

	// append config-backup
	QDomElement config_backup = xml.createElementNS(CAPABILITY_MAINTAIN_BASE, "config-backup");
	root.appendChild(config_backup);

	return xml.toString();
}

void StoreBackup::talk()
{
	emit this->writeMessage(StoreBackup::storeBackupMessage(this->rpcMessageId));
}


QList<Capability> RestoreBackup::requiredCapabilities = QList<Capability>() << Capability(CAPABILITY_MAINTAIN_BASE, QString());

void RestoreBackup::readMessage(QString msg)
{
	// Parse xml
	QDomDocument xml;
	if (!this->parseXml(xml, msg))
		return;

	// Check message id
	if (!this->checkRpcMessageId(xml))
		return;

	bool res = true;
	// Look for rpc-error
	QList<NetconfError> errors;
	if (this->parseNetconfErrors(xml, errors)) {
		emit this->netconfErrorsOccured(errors);
		emit this->report(Communicator::RT_ERROR, tr("Failed to restore the backup."));
		res = false;
	} else {
		QDomNodeList ips = xml.elementsByTagName("new-ip");
		if (ips.length() == 1) {
			emit this->dataRestored(ips.at(0).toElement().text());
		} else {
			// new-ip tag is not mandatory
			emit this->dataRestored(QString());
		}
	}

	this->finish(res);
}

QString RestoreBackup::restoreBackupMessage(int rpcMessageId, QByteArray &data)
{
	QDomDocument xml;

	QDomElement root = buildXmlHeaderAndRpc(xml, rpcMessageId);

	// append config-backup
	QDomElement config_backup = xml.createElementNS(CAPABILITY_MAINTAIN_BASE, "config-backup");
	root.appendChild(config_backup);

	// append data
	QDomElement data_element = xml.createElement("data");
	data_element.appendChild(xml.createTextNode(QString(data.toBase64())));
	config_backup.appendChild(data_element);

	return xml.toString();
}

void RestoreBackup::talk()
{
	emit this->writeMessage(RestoreBackup::restoreBackupMessage(this->rpcMessageId, this->data));
}

QList<Capability> GetStats::requiredCapabilities = QList<Capability>() << Capability(CAPABILITY_STATS_BASE, QString());

void GetStats::readMessage(QString msg)
{
	// Parse xml
	QDomDocument xml;
	if (!this->parseXml(xml, msg))
		return;

	// Check message id
	if (!this->checkRpcMessageId(xml))
		return;


	QDomNodeList stats = xml.elementsByTagName("stats");
	if (stats.count() != 1) {
		emit this->report(Communicator::RT_ERROR, tr("Incorrect Stat XML format."));
		this->finish(false);
		return;

	}

	StatsRecord record;
	record.model = stats.at(0).firstChildElement("model").text();
	record.boardName = stats.at(0).firstChildElement("board-name").text();
	record.kernelVersion= stats.at(0).firstChildElement("kernel-version").text();
	record.osVersion = stats.at(0).firstChildElement("turris-os-version").text();
	record.uptime = stats.at(0).firstChildElement("uptime").text().toFloat();
	record.hostname = stats.at(0).firstChildElement("hostname").text();
	QDomElement firewallSending = stats.at(0).firstChildElement("firewall-sending");
	record.firewallStatus = firewallSending.firstChildElement("status").text();
	record.firewallAge = firewallSending.firstChildElement("age").text().toInt();
	QDomElement ucollectSending = stats.at(0).firstChildElement("ucollect-sending");
	record.ucollectStatus = ucollectSending.firstChildElement("status").text();
	record.ucollectAge = ucollectSending.firstChildElement("age").text().toInt();

	QDomElement interface = stats.at(0).firstChildElement("interfaces").firstChildElement("interface");
	while (!interface.isNull()) {
		StatsInterface interfaceRecord;
		interfaceRecord.down = interface.firstChildElement("down").isNull() ? false : true;
		interfaceRecord.up = interface.firstChildElement("up").isNull() ? false : true;
		interfaceRecord.use = interface.firstChildElement("use").text();
		interfaceRecord.name = interface.firstChildElement("name").text();
		QDomElement e = interface.firstChildElement("address");
		while (!e.isNull()) {
			StatsAddress statsAddress;
			statsAddress.address = e.firstChildElement("address").text();
			statsAddress.type = e.firstChildElement("type").text();
			e = e.nextSiblingElement("address");
			interfaceRecord.addresses.append(statsAddress);
		}
		record.interfaces.append(interfaceRecord);
		interface = interface.nextSiblingElement("interface");
	}

	emit this->exportStats(record);

	this->finish(true);
}

QString GetStats::getStatsMessage(int rpcMessageId)
{
	QDomDocument xml;

	QDomElement root = buildXmlHeaderAndRpc(xml, rpcMessageId);

	// append get
	QDomElement get = xml.createElement("get");
	root.appendChild(get);

	// append filter
	QDomElement filter = xml.createElement("filter");
	filter.setAttribute("type", "subtree");
	get.appendChild(filter);

	// append stats
	QDomElement stats = xml.createElementNS(CAPABILITY_STATS_BASE, "stats");
	filter.appendChild(stats);

	return xml.toString();
}

void GetStats::talk()
{
	emit this->writeMessage(GetStats::getStatsMessage(this->rpcMessageId));
}


QList<Capability> GetSerial::requiredCapabilities = QList<Capability>() << Capability(CAPABILITY_REGISTRATION_BASE, QString());

void GetSerial::readMessage(QString msg)
{
	// Parse xml
	QDomDocument xml;
	if (!this->parseXml(xml, msg))
		return;

	// Check message id
	if (!this->checkRpcMessageId(xml))
		return;

	// Export serial
	bool res;
	QDomNodeList serials = xml.elementsByTagName("serial");
	if (serials.length() == 1) {
		emit this->exportSerial(serials.at(0).toElement().text());
		res = true;
	} else {
		res = false;
	}

	this->finish(res);
}

QString GetSerial::getSerialMessage(int rpcMessageId)
{
	QDomDocument xml;

	QDomElement root = buildXmlHeaderAndRpc(xml, rpcMessageId);

	// append serial
	QDomElement serial = xml.createElementNS(CAPABILITY_REGISTRATION_BASE, "serial");
	root.appendChild(serial);

	return xml.toString();
}

void GetSerial::talk()
{
	emit this->writeMessage(GetSerial::getSerialMessage(this->rpcMessageId));
}


QList<Capability> GetUpdaterData::requiredCapabilities = QList<Capability>() << Capability(CAPABILITY_UPDATER_BASE, QString());

void GetUpdaterData::readMessage(QString msg)
{
	// Parse xml
	QDomDocument xml;
	if (!this->parseXml(xml, msg))
		return;

	// Check message id
	if (!this->checkRpcMessageId(xml))
		return;

	QDomNodeList updater = xml.elementsByTagName("updater");
	if (updater.count() != 1) {
		emit this->report(Communicator::RT_ERROR, tr("Incorrect updater XML format."));
		this->finish(false);
		return;
	}

	UpdaterDataRecord record;
	record.status = "inactive";
	QDomElement running = updater.at(0).firstChildElement("running");
	if (!running.isNull()) {
		record.status = running.text();
	}

	QList<PkgListRecord> userListRecords;
	QDomNodeList userLists = xml.elementsByTagName("pkg-list");
	for (int i = 0; i < userLists.count(); ++i) {
		PkgListRecord pkgRecord;
		pkgRecord.name = userLists.at(i).firstChildElement("name").text();

		QString noLangText;
		// Load multilang title
		QDomElement titleTextElement = userLists.at(i).firstChildElement("title");
		while (!titleTextElement.isNull()) {
			if (titleTextElement.attribute("lang") == this->lang) {
				pkgRecord.title = titleTextElement.text();
				break;
			} else {
				noLangText = titleTextElement.text();
			}
			titleTextElement = titleTextElement.nextSiblingElement("title");
		}
		if (pkgRecord.title.isNull())
			pkgRecord.title = noLangText;

		// Load multilang description
		QDomElement descriptionTextElement = userLists.at(i).firstChildElement("description");
		while (!descriptionTextElement.isNull()) {
			if (descriptionTextElement.attribute("lang") == this->lang) {
				pkgRecord.description= descriptionTextElement.text();
				break;
			} else {
				noLangText = descriptionTextElement.text();
			}
			descriptionTextElement = descriptionTextElement.nextSiblingElement("description");
		}
		if (pkgRecord.description.isNull())
			pkgRecord.description = noLangText;

		// Get activated flag
		QDomElement activated = userLists.at(i).firstChildElement("activated");
		pkgRecord.activated = !activated.isNull();

		userListRecords.append(pkgRecord);
	}

	QDomNodeList lastActivity = xml.elementsByTagName("last_activity");
	if (lastActivity.count() != 1) {
		emit this->report(Communicator::RT_ERROR, tr("Incorrect Stat XML format."));
		this->finish(false);
		return;
	}
	QDomNodeList activities = lastActivity.at(0).childNodes();
	for (int i = 0; i < activities.count(); ++i) {
		QDomElement activity = activities.at(i).toElement();
		QString packageName, version;
		foreach (QString str, activity.text().split(" ")) {
			if (packageName.isEmpty()) {
				packageName = str;
			} else {
				version += str;
			}
		}
		UpdaterActivityRecord activityRecord;
		activityRecord.action = activity.tagName();
		activityRecord.packageName = packageName;
		activityRecord.packageVersion = version;
		record.lastActivities.append(activityRecord);
	}

	record.userLists = userListRecords;
	emit this->exportLists(record);

	this->finish(true);
}

QString GetUpdaterData::getUpdaterDataMessage(int rpcMessageId)
{
	QDomDocument xml;

	QDomElement root = buildXmlHeaderAndRpc(xml, rpcMessageId);

	// append get
	QDomElement get = xml.createElement("get");
	root.appendChild(get);

	// append filter
	QDomElement filter = xml.createElement("filter");
	filter.setAttribute("type", "subtree");
	get.appendChild(filter);

	// append updater
	QDomElement updater = xml.createElementNS(CAPABILITY_UPDATER_BASE, "updater");
	filter.appendChild(updater);

	return xml.toString();
}

void GetUpdaterData::talk()
{
	emit this->writeMessage(GetUpdaterData::getUpdaterDataMessage(this->rpcMessageId));
}


QList<Capability> CheckUpdater::requiredCapabilities = QList<Capability>() << Capability(CAPABILITY_UPDATER_BASE, QString());

void CheckUpdater::readMessage(QString msg) {
	// Parse xml
	QDomDocument xml;
	if (!this->parseXml(xml, msg))
		return;

	// Check message id
	if (!this->checkRpcMessageId(xml))
		return;

	// Check Ok
	bool res = this->checkRpcMessageOk(xml, false);

	this->finish(res);
}

QString CheckUpdater::checkUpdaterMessage(int rpcMessageId)
{
	QDomDocument xml;

	QDomElement root = buildXmlHeaderAndRpc(xml, rpcMessageId);

	// append display
	QDomElement display = xml.createElementNS(CAPABILITY_UPDATER_BASE, "check");
	root.appendChild(display);

	return xml.toString();
}

void CheckUpdater::talk()
{
	emit this->writeMessage(CheckUpdater::checkUpdaterMessage(this->rpcMessageId));
}


QList<Capability> SetUpdater::requiredCapabilities = QList<Capability>() << Capability(CAPABILITY_UPDATER_BASE, QString());

void SetUpdater::readMessage(QString msg) {
	// Parse xml
	QDomDocument xml;
	if (!this->parseXml(xml, msg))
		return;

	// Check message id
	if (!this->checkRpcMessageId(xml))
		return;

	// Check Ok
	bool res = this->checkRpcMessageOk(xml, false);
	if (!res) {
		QList<NetconfError> errors;
		if (this->parseNetconfErrors(xml, errors))
			emit this->netconfErrorsOccured(errors);
	}

	this->finish(res);
}

QString SetUpdater::setUpdaterMessage(int rpcMessageId, QList<QString> newList)
{
	QDomDocument xml;

	QDomElement root = buildXmlHeaderAndRpc(xml, rpcMessageId);

	// append edit-config
	QDomElement edit_config = xml.createElement("edit-config");
	root.appendChild(edit_config);

	// append target
	QDomElement target = xml.createElement("target");
	target.appendChild(xml.createElement("running"));
	edit_config.appendChild(target);

	// append config
	QDomElement config = xml.createElement("config");
	edit_config.appendChild(config);

	// append updater-config
	QDomElement updater_config = xml.createElementNS(CAPABILITY_UPDATER_BASE, "updater-config");
	config.appendChild(updater_config);

	// append active_lists
	QDomElement active_lists = xml.createElement("active-lists");
	//active_lists.setPrefix("nc");
	QDomAttr attr_nc = xml.createAttribute("xmlns:nc");
	attr_nc.setValue("urn:ietf:params:xml:ns:netconf:base:1.0");
	QDomAttr attr_replace = xml.createAttribute("nc:operation");
	attr_replace.setValue("replace");
	active_lists.setAttributeNode(attr_nc);
	active_lists.setAttributeNode(attr_replace);
	updater_config.appendChild(active_lists);

	foreach (QString listName, newList) {
		// append user-list
		QDomElement user_list = xml.createElement("user-list");
		active_lists.appendChild(user_list);

		// append name
		QDomElement name_element = xml.createElement("name");
		name_element.appendChild(xml.createTextNode(listName));
		user_list.appendChild(name_element);
	}

	return xml.toString();
}

void SetUpdater::talk()
{
	emit this->writeMessage(SetUpdater::setUpdaterMessage(this->rpcMessageId, this->newList));
}


QList<Capability> GetNeighbours::requiredCapabilities = QList<Capability>() << Capability(CAPABILITY_NEIGHBOURS_BASE, QString());

void GetNeighbours::readMessage(QString msg) {
	// Parse xml
	QDomDocument xml;
	if (!this->parseXml(xml, msg))
		return;

	// Check message id
	if (!this->checkRpcMessageId(xml))
		return;

	// read neighbours
	QDomNodeList neighbours = xml.elementsByTagName("neighbour");
	if (!neighbours.isEmpty()) {
		QList<NeighbourRecord> result;
		for (int i = 0; i < neighbours.size(); ++i) {
			NeighbourRecord neighbourRecord;
			// read mac address
			QDomElement macAddress = neighbours.at(i).firstChildElement("mac-address");
			neighbourRecord.macAddress = macAddress.isNull() ? QString() : macAddress.text();

			// read interface
			QDomElement interface = neighbours.at(i).firstChildElement("interface");
			neighbourRecord.interfaceName = interface.isNull() ? QString() : interface.text();

			// read ips
			QDomElement e = neighbours.at(i).firstChildElement("ip-address");
			while (!e.isNull()) {
				NeighbourIpRecord ipRecord;

				QDomElement ip = e.firstChildElement("ip");
				ipRecord.ipAddress = ip.isNull() ? "" : ip.text();
				QDomElement count = e.firstChildElement("connection-count");
				ipRecord.connectionCount = count.isNull() ? 0 : count.text().toInt();
				QDomElement nud = e.firstChildElement("nud");
				ipRecord.nud = nud.isNull() ? QString() : nud.text();
				QDomElement lease = e.firstChildElement("dhcp-lease");
				ipRecord.dhcpLease = lease.isNull() ? QDateTime() : QDateTime::fromTime_t(lease.text().toInt());
				QDomElement hostname = e.firstChildElement("hostname");
				ipRecord.hostname = hostname.isNull() ? QString() : hostname.text();

				QDomElement statistics = e.firstChildElement("statistics");
				if (!statistics.isNull()) {
					QDomElement used = statistics.firstChildElement("used");
					ipRecord.usedSince = used.isNull() ? -1 : used.text().toInt();
					QDomElement confirmed = statistics.firstChildElement("confirmed");
					ipRecord.confirmedSince = confirmed.isNull() ? -1 : confirmed.text().toInt();
					QDomElement updated = statistics.firstChildElement("updated");
					ipRecord.updatedSince = updated.isNull() ? -1 : updated.text().toInt();
				}

				e = e.nextSiblingElement("ip-address");
				neighbourRecord.ips.append(ipRecord);
			}
			result.append(neighbourRecord);
		}
		emit this->exportNeighbours(result);
	} else {
		emit this->exportNeighbours(QList<NeighbourRecord>());
	}

	this->finish(true);
}

QString GetNeighbours::getNeighbourMessage(int rpcMessageId, QList<QString> interfaces)
{
	QDomDocument xml;

	QDomElement root = buildXmlHeaderAndRpc(xml, rpcMessageId);

	// append get
	QDomElement get = xml.createElement("get");
	root.appendChild(get);

	// append filter
	QDomElement filter = xml.createElement("filter");
	filter.setAttribute("type", "subtree");
	get.appendChild(filter);

	QDomElement messages = xml.createElementNS(CAPABILITY_NEIGHBOURS_BASE, "neighbours");
	filter.appendChild(messages);

	foreach (QString interface, interfaces) {
		QDomElement neighbour = xml.createElement("neighbour");
		messages.appendChild(neighbour);
		QDomElement interfaceDom = xml.createElement("interface");
		neighbour.appendChild(interfaceDom);
		interfaceDom.appendChild(xml.createTextNode(interface));
	}

	return xml.toString();
}

void GetNeighbours::talk()
{
	emit this->writeMessage(GetNeighbours::getNeighbourMessage(this->rpcMessageId, this->interfaces));
}
