#include "communicator.h"
#include "logging.h"

#include <QFile>
#include <QSslKey>
#include <QRegExp>
#include <QSslCipher>

#define MSG_TERMINATOR "]]>]]>"

Connector::Connector(const QString &server, int port, const QString &tokenPath) :
		socket(NULL),
		server(server),
		port(port),
		tokenPath(tokenPath),
		state(Communicator::CONNECTOR_STARTED) {}

void Connector::initiateConnection()
{
	// load configuration
	if (this->createConfiguration(this->tokenPath)) {
		this->state = Communicator::CONNECTOR_CONFIGURED;
	} else {
		this->state = Communicator::CONNECTOR_ERROR;
		return;
	}


	this->socket = new QSslSocket(this);
	this->socket->setSslConfiguration(*this->configuration);
	this->socket->setSocketOption(QAbstractSocket::KeepAliveOption, 1);

	//this->socket->setCiphers("HIGH:!LOW:!MEDIUM:!SSLv2:!aNULL:!eNULL:!DES:!3DES:!AES128:!CAMELLIA128");

	// connect signals
	QObject::connect(this->socket, SIGNAL(error(QAbstractSocket::SocketError)), this, SLOT(handleSocketError(QAbstractSocket::SocketError)));
	QObject::connect(this->socket, SIGNAL(stateChanged(QAbstractSocket::SocketState)), this, SLOT(handleSocketStateChange(QAbstractSocket::SocketState)));
	QObject::connect(this->socket, SIGNAL(connected()), this, SLOT(handleSocketConnected()));
	QObject::connect(this->socket, SIGNAL(disconnected()), this, SLOT(handleSocketDisconnected()));
	QObject::connect(this->socket, SIGNAL(sslErrors(QList<QSslError>)), this, SLOT(handleSslErrors(QList<QSslError>)));
	QObject::connect(this->socket, SIGNAL(readyRead()), this, SLOT(handleReadReady()));
	QObject::connect(this->socket, SIGNAL(encrypted()), this, SLOT(handleEncrypted()));
	QObject::connect(&this->chatTimer, SIGNAL(timeout()), this, SLOT(handleChatTimeOut()));

	if (Communicator::CONNECTOR_CONFIGURED == this->state) {
		log_connection.debug() << "Starting to connect...";
		this->socket->connectToHostEncrypted(server, port);
	}
}

bool Connector::createConfiguration(const QString &tokenPath)
{
	this->configuration = new QSslConfiguration();

	// read the whole file and parse the certificates
	QFile tokenFile(tokenPath);
	if(!tokenFile.open(QIODevice::ReadOnly))
	{
		emit this->report(Communicator::RT_ERROR, tr("Failed to open the token file."));
		log_connection.critical() << "Failed to open the token file:" << tokenFile.errorString();
		return false;
	}
	QString fileContent = tokenFile.readAll();
	tokenFile.close();

	QStringList list = fileContent.split(QRegExp("-----BEGIN"));
	if (list.size() != 4) {
		emit this->report(Communicator::RT_ERROR, tr("Failed to read the token file."));
		log_connection.critical() << "Failed to read the token file. The file has an incorrect format.";
		return false;
	}

	QByteArray certData(("-----BEGIN" + list.at(1)).toStdString().c_str());
	QByteArray keyData(("-----BEGIN" + list.at(2)).toStdString().c_str());
	QByteArray CAData(("-----BEGIN" + list.at(3)).toStdString().c_str());

	QSslCertificate clientCert(certData);
	if (clientCert.isNull()) {
		emit this->report(Communicator::RT_ERROR, tr("Failed to load client certificate."));
		log_connection.critical() << "Failed to load client certificate.";
		return false;
	}
#ifdef MORE_DEBUG
	log_connection.debug() << "Client certificate: \n" << clientCert.toPem();
#else
	log_connection.debug() << "Client certificate:" << clientCert.toPem().length() << "characters";
#endif

	QSslKey clientKey(keyData, QSsl::Rsa);
	if (clientKey.isNull()) {
		emit this->report(Communicator::RT_ERROR, tr("Failed to load client key."));
		log_connection.critical() << "Failed to load client key.";
		return false;
	}
#ifdef MORE_DEBUG
	log_connection.debug() << "Client key: \n" << clientKey.toPem();
#else
	log_connection.debug() << "Client key:" << clientKey.toPem().length() << "characters";
#endif

	QSslCertificate CACert(CAData);
	if (CACert.isNull()) {
		emit this->report(Communicator::RT_ERROR, tr("Failed to load CA certificate."));
		log_connection.critical() << "Failed to load CA certificate.";
		return false;
	}
#ifdef MORE_DEBUG
	log_connection.debug() << "CA certificate: \n" << CACert.toPem();
#else
	log_connection.debug() << "CA certificate:" << CACert.toPem().length() << "characters";
#endif

	this->configuration->setProtocol(QSsl::SecureProtocols);
	this->configuration->setLocalCertificate(clientCert);
	this->configuration->setPrivateKey(clientKey);
	this->configuration->setCaCertificates(QList<QSslCertificate>() << CACert);

	return true;
}

void Connector::disposeConfiguration()
{
	if (this->configuration)
		delete this->configuration;
	this->configuration = NULL;
}

void Connector::connectChat(Chats::Base *chat)
{
	QObject::connect(this, SIGNAL(readMessage(QString)), chat, SLOT(readMessage(QString)));
	QObject::connect(chat, SIGNAL(finished(bool)), this, SLOT(handleChatFinished(bool)));
	QObject::connect(chat, SIGNAL(writeMessage(QString)), this, SLOT(writeMessage(QString)));
	QObject::connect(chat, SIGNAL(report(Communicator::ReportType, QString)), this, SLOT(handleReport(Communicator::ReportType,QString)));
	QObject::connect(chat, SIGNAL(communicationStatusUpdate(Communicator::ConnectorStates)), this, SLOT(handleStateChange(Communicator::ConnectorStates)));
	QObject::connect(chat, SIGNAL(netconfErrorsOccured(QList<NetconfError>)), this, SLOT(handleNetconfErrors(QList<NetconfError>)));
}

bool Connector::startChat(Chats::Base *chat, bool requireCapabilities)
{
	if (!this->capabilityCheck(chat)) {
		if (requireCapabilities) {
			log_connection.debug() << "Chat won't be started due to it lacks some of the requirements.";
			return false;
		} else {
			log_connection.debug() << "Chat will be started although it lacks some of the requirements.";
		}
	}
	log_connection.debug() << "Starting Chat" << chat->getName();
	this->connectChat(chat);
	this->resetChatTimer();
	chat->talk();
	return true;
}

void Connector::performInitialChat(Chats::Initial *chat)
{
	QObject::connect(chat, SIGNAL(exportCapabilities(QList<Capability>)), this, SLOT(handleStoreCapabilities(QList<Capability>)));
	this->performChat(chat, Communicator::CONNECTOR_ANY, false);
}

void Connector::performChat(Chats::Base *chat, Communicator::ConnectorStates requiredState, bool requireCapabilities)
{
	if (this->chatQueue.empty()) {
		if (requiredState == Communicator::CONNECTOR_ANY || this->state == requiredState) {
			this->chatQueue.append(new QueuedChat(chat, requiredState, requireCapabilities));
			// start immediately
			if (!this->startChat(chat, requireCapabilities)) {
				// chat fails to start it needs to be removed from the queue
				this->popChatQueue();
			}
		} else {
			emit this->report(Communicator::RT_ERROR, tr("Incorrect communication state."));
			emit this->chatRemoved(chat);
		}
	} else {
		log_connection.debug() << "Queued Chat" << chat->getName();
		this->chatQueue.push_back(new QueuedChat(chat, requiredState, requireCapabilities));
	}
}

void Connector::resetChatTimer()
{
	this->chatTimer.stop();
	this->chatTimer.start(CHAT_TIMEOUT);
}

void Connector::popChatQueue() {
	if (!this->chatQueue.isEmpty()) {
		QueuedChat *qchat = this->chatQueue[0];
		this->chatQueue.pop_front();
		emit this->chatRemoved(qchat->chat);
		delete qchat;
	}
}

void Connector::writeMessage(QString msg)
{
#ifdef MORE_DEBUG
	log_connection.debug() << "Sending a message:\n" << msg;
#else
	log_connection.debug() << "Sending a message:" << msg.length() << "characters";
#endif
	msg += MSG_TERMINATOR "\n";
	if (this->socket->state() == QAbstractSocket::ConnectedState && this->socket->isEncrypted()) {
		if (0 > this->socket->write(QByteArray(msg.toStdString().c_str()))) {
			log_connection.critical() << "Failed to write into the socket.";
		}
	} else {
		log_connection.critical() << "Can write messge, socket not connected.";
	}
}

void Connector::startNextChat()
{
	this->popChatQueue();

	while (!this->chatQueue.empty()) {
		QueuedChat *qchat = this->chatQueue[0];
		if (this->state == qchat->state) {
			// start queued chat
			if (!this->startChat(qchat->chat, qchat->capabilityRequired)) {
				// chat fails to start it needs to be removed from the queue
				this->popChatQueue();
			} else {
				// chat successfully started
				break;
			}
		} else {
			emit this->report(Communicator::RT_ERROR, tr("Incorrect communication state."));
			this->popChatQueue();
		}
	}
}

void Connector::handleChatFinished(bool success)
{
	this->chatTimer.stop();
	log_connection.debug() << "Chat finished" << (success ? "successfully.": "unsuccessfully.");
	this->startNextChat();
}

Connector::~Connector()
{
	log_connection.debug() << "Closing connection...";
	this->popChatQueue();
	if (this->socket) {
		if (this->socket->isOpen())
			this->socket->close();
		delete this->socket;
	}
	for (QList<QueuedChat *>::iterator i = this->chatQueue.begin(); i != this->chatQueue.end(); ++i) {
		emit this->chatRemoved((*i)->chat);
		delete (*i);
	}
	this->disposeConfiguration();
}

void Connector::handleSocketError(QAbstractSocket::SocketError error)
{
	Q_UNUSED(error)
	emit this->report(Communicator::RT_ERROR, tr("Socket error occured."));
	log_connection.critical() << "Socket Error:" << this->socket->errorString();
	this->state = Communicator::CONNECTOR_ERROR;
	emit this->disconnect();
}

void Connector::handleSslErrors(const QList<QSslError> &errors)
{
	emit this->report(Communicator::RT_ERROR, tr("SSL error occured."));
	log_connection.critical() << "SSL Errors:";
	QList<QSslError>::const_iterator i;
	for (i = errors.begin(); i != errors.end(); ++i) {
		log_connection.critical() << (int)i->error() << i->errorString();
		if (i->error() == QSslError::HostNameMismatch) {
			this->socket->ignoreSslErrors(QList<QSslError>() << *i);
			return;
		}
	}
	this->state = Communicator::CONNECTOR_ERROR;
	emit this->disconnect();
}

void Connector::handleSocketStateChange(QAbstractSocket::SocketState state)
{
	log_connection.debug() << "New socket state:" << state;
}

void Connector::handleStateChange(Communicator::ConnectorStates newState)
{
	log_connection.debug() << "New communicator state:" << newState;
	this->state = newState;
	if (Communicator::CONNECTOR_READY) {
		emit connected();
	}
}

void Connector::handleSocketConnected()
{
	log_connection.debug() << "Socket Connected.";
}

void Connector::handleSocketDisconnected()
{
	emit this->report(Communicator::RT_SUCCESS, tr("Disconnected from the router."));
	log_connection.debug() << "Socket Disconnected.";
}

void Connector::handleReadReady()
{
	this->readBuffer.append(this->socket->readAll());
	if (this->readBuffer.contains(MSG_TERMINATOR)) {
		QString msg(this->readBuffer);
		msg.replace(QRegExp(QRegExp::escape(MSG_TERMINATOR)), "");
#ifdef MORE_DEBUG
		log_connection.debug() << "Recived a message:\n" << msg;
#else
		log_connection.debug() << "Recived a message:" << msg.length() << "characters";
#endif
		emit this->readMessage(msg);
		this->readBuffer.clear();
	}
}

void Connector::handleEncrypted()
{
	emit this->report(Communicator::RT_SUCCESS, tr("Connected to the router."));
	log_connection.debug() << "Socket Encrypted.";
}

void Connector::handleReport(Communicator::ReportType type, QString msg)
{
	emit this->report(type, msg);
}

void Connector::handleChatTimeOut()
{
	if (Communicator::CONNECTOR_TIMEOUT != this->state) {
		this->state = Communicator::CONNECTOR_TIMEOUT;
		emit this->disconnected();
	}
}

void Connector::handleStoreCapabilities(QList<Capability> capabilities)
{
	this->capabilities = capabilities;
	foreach (Capability capability, this->capabilities) {
		log_connection.debug() << "capability obtained" << capability.base << capability.revision;
	}
}

void Connector::handleNetconfErrors(QList<NetconfError> errors)
{
	foreach (NetconfError error, errors) {
		QString msg = "Netconf error occured:\n";
		msg += "type: " + error.type + "\n";
		msg += "tag: " + error.tag + "\n";
		msg += "severity: " + error.serverity + "\n";
		msg += "app-tag: " + error.appTag + "\n";
		msg += "path: " + error.path + "\n";
		msg += "message: " + error.message + "\n";
		msg += "info: \n";
		foreach (NetconfErrorInfoElement iError, error.info) {
			msg += "  " + iError.name + ": " + iError.content + "\n";
		}
		log_connection.critical() << msg.toStdString().c_str();
	}
}

bool Connector::ready()
{
	return Communicator::CONNECTOR_READY == this->state;
}

bool Connector::capabilityCheck(Chats::Base *chat)
{
	bool result = true;
	// Check whether the connector has all required capabilities
	foreach (Capability requiredCapability, chat->getRequiredCapabilities()) {
		bool has = false;
		foreach (Capability ownedCapability, this->capabilities) {
			if (ownedCapability.covers(requiredCapability)) {
				has = true;
				break;
			}
		}
		if (has) {
			log_connection.debug() << "server has required capability" << requiredCapability.base << requiredCapability.revision << "for chat" << chat->getName();
		} else {
			log_connection.warning() << "server is missing capability" << requiredCapability.base << requiredCapability.revision << "for chat" << chat->getName();
			result = false;
		}
	}
	return result;
}
