/*
 * Spectator - a desktop app which communicates with router Turris using NETCONF protocol
 * Copyright (C) 2015  CZ.NIC, z. s. p. o. <http://www.nic.cz>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef SPECTATOR_COMMUNICATOR_H
#define SPECTATOR_COMMUNICATOR_H

#include <QObject>
#include <QHash>
#include <QList>
#include <QString>
#include <QtNetwork/qsslsocket.h>
#include <QSslConfiguration>

#include "conn_consts.h"
#include "chat.h"

#define CHAT_TIMEOUT 5000  // 5 seconds

struct QueuedChat {
	Chats::Base *chat;
	Communicator::ConnectorStates state;
	bool capabilityRequired;

	QueuedChat(Chats::Base *chat, Communicator::ConnectorStates state, bool capabilityRequired = true)
		: chat(chat), state(state), capabilityRequired(capabilityRequired) {}
};

class Connector : public QObject {
	Q_OBJECT

public:
	QList<Capability> capabilities;
	bool ready();

	void performInitialChat(Chats::Initial *chat);
	Connector(const QString &server, int port, const QString &tokenPath);
	~Connector();

public slots:
	void initiateConnection();
	void performChat(
		Chats::Base *chat,
		Communicator::ConnectorStates requiredState = Communicator::CONNECTOR_READY,
		bool requireCapabilities = true
	);
	void writeMessage(QString msg);

private:
	QSslSocket *socket;
	QSslConfiguration *configuration;
	QString server;
	int port;
	QString tokenPath;
	QByteArray readBuffer;
	QList<QueuedChat *> chatQueue;
	Communicator::ConnectorStates state;
	QTimer chatTimer;

	bool createConfiguration(const QString &tokenPath);
	void disposeConfiguration();
	void connectChat(Chats::Base *chat);
	void popChatQueue();
	void resetChatTimer();
	bool capabilityCheck(Chats::Base *chat);
	bool startChat(Chats::Base *chat, bool requireCapabilities);
	void startNextChat();

signals:
	void chatRemoved(Chats::Base *chat);
	void report(Communicator::ReportType type, QString msg);
	void readMessage(QString msg);
	void connected();
	void disconnected();

private slots:
	void handleChatFinished(bool);
	void handleSocketError(QAbstractSocket::SocketError error);
	void handleSslErrors(const QList<QSslError> &errors);
	void handleSocketStateChange(QAbstractSocket::SocketState socketState);
	void handleStateChange(Communicator::ConnectorStates newState);
	void handleSocketConnected();
	void handleSocketDisconnected();
	void handleReadReady();
	void handleEncrypted();
	void handleReport(Communicator::ReportType type, QString msg);
	void handleChatTimeOut();
	void handleStoreCapabilities(QList<Capability> capabilities);
	void handleNetconfErrors(QList<NetconfError> errors);
};

#endif
