/*
 * Spectator - a desktop app which communicates with router Turris using NETCONF protocol
 * Copyright (C) 2015  CZ.NIC, z. s. p. o. <http://www.nic.cz>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "logging.h"
#include "settings.h"
#include "chat.h"

#include "communicator.h"

#define RECONNECT_TIMEOUT 10000  // 10s

#define TAB_IDX_NOTIFICATION 0
#define TAB_IDX_NETHIST 1
#define TAB_IDX_STATS 2
#define TAB_IDX_UPDATER 3
#define TAB_IDX_NEIGHBOURS 4

static inline void setTextLabel(QLabel *label, QString newText)
{
	if (newText.isEmpty()) {
		label->setText("???");
		label->setToolTip(QString());
	} else {
		label->setText(newText);
		label->setToolTip(newText);
	}
}

MainWindow::MainWindow(QWidget *parent, QSystemTrayIcon *trayIcon) :
	QMainWindow(parent),
	settings(NULL),
	trayIcon(trayIcon),
	ui(new Ui::MainWindow),
	connector(NULL),
	enableNeighbourChat(true)
{
	// Set required capabilities
	this->tabCapabilities[TAB_IDX_NOTIFICATION] = QList<Capability>()
			<< Chats::GetNotifications::requiredCapabilities
			<< Chats::DeleteNotification::requiredCapabilities;
	this->tabCapabilities[TAB_IDX_NETHIST] = QList<Capability>()
			<< Chats::GetNethist::requiredCapabilities;
	this->tabCapabilities[TAB_IDX_STATS] = QList<Capability>()
			<< Chats::SetPassword::requiredCapabilities
			<< Chats::GetSerial::requiredCapabilities
			<< Chats::GetStats::requiredCapabilities;
	this->tabCapabilities[TAB_IDX_UPDATER] = QList<Capability>()
			<< Chats::CheckUpdater::requiredCapabilities
			<< Chats::GetUpdaterData::requiredCapabilities
			<< Chats::SetUpdater::requiredCapabilities;
	this->tabCapabilities[TAB_IDX_NEIGHBOURS] = QList<Capability>()
			<< Chats::GetNeighbours::requiredCapabilities;


	// Setup the gui
	this->ui->setupUi(this);
	this->ui->rebootDeviceButton->setIcon(QApplication::style()->standardIcon(QStyle::SP_BrowserReload));
	this->createMenus();

	// Set tab action list (has to be called after createMenus)
	this->tabActions = QList<QAction *>()
			<< this->notificationAction << this->statisticsAction << this->deviceAction
			<< this->updaterAction << this->neighboursAction;

	this->interfaceWidget = new InterfaceWidget(this->ui->statisticsNetworkFrame, this);

	this->updaterUserList = new UpdaterUserListsWidget(this, this);
	this->ui->updaterScrollArea->setWidget(this->updaterUserList);
	this->updaterActivities = new UpdaterActivitiesWidget(this);
	this->ui->updaterLogScrollArea->setWidget(this->updaterActivities);

	// Set refresh timer signal
	QObject::connect(&this->refreshTimer, SIGNAL(timeout()), this, SLOT(callPerformLoopChats()));
	// Set reconnect timer signal
	QObject::connect(&this->reconnectTimer, SIGNAL(timeout()), this, SLOT(handleReconnect()));
	// Handle tab switching
	QObject::connect(this->ui->mainTab, SIGNAL(currentChanged(int)), this, SLOT(callPerformLoopChats()));
	QObject::connect(this->ui->mainTab, SIGNAL(currentChanged(int)), this, SLOT(callNeighbourMacEditDisable()));

	// neighbour advanced view switch
	QObject::connect(this->ui->neighboursAdvancedViewCheckBox, SIGNAL(toggled(bool)), this, SLOT(handleNeighbourAdvancedViewToggled(bool)));

	this->initCharts();
}

void MainWindow::contextMenuEvent(QContextMenuEvent *event)
{
	this->menu->exec(event->globalPos());
}

void MainWindow::changeEvent(QEvent* event)
{
	if (event->type() == QEvent::LanguageChange)
	{
		this->ui->retranslateUi(this);
		// retranslate menu
		this->notificationAction->setText(tr("Notifications"));
		this->settingsAction->setText(tr("Settings"));
		this->deviceAction->setText(tr("Device"));
		this->statisticsAction->setText(tr("Statistics"));
		this->updaterAction->setText(tr("Updater"));
		this->neighboursAction->setText(tr("Neighbours"));
		this->aboutAction->setText(tr("About"));
		this->quitAction->setText(tr("Quit"));

		this->interfaceWidget->updateTexts();
		this->updaterUserList->updateTexts();
	}

	QMainWindow::changeEvent(event);
}

void MainWindow::keyPressEvent(QKeyEvent *event)
{
	switch (event->key()) {
		case Qt::Key_Escape:
			if (this->isVisible() && this->isActiveWindow() && this->isInSystrayMode()) {
				this->toggleWindow(QSystemTrayIcon::Unknown);
			}
			// Close neighbour mac edits
			this->callNeighbourMacEditDisable();
			break;
	}
}

MainWindow::~MainWindow()
{
	delete this->connector;
	delete this->ui;
	delete this->menu;
	delete this->neighbourInterfaceMenu;
	delete this->quitAction;
	delete this->settingsAction;
	delete this->notificationAction;
	delete this->statisticsAction;
	delete this->interfaceWidget;
}

void MainWindow::createMenus()
{
	// Init menu actions
	this->notificationAction = new QAction(tr("Notifications"), this);
	this->settingsAction = new QAction(tr("Settings"), this);
	this->deviceAction = new QAction(tr("Device"), this);
	this->statisticsAction= new QAction(tr("Statistics"), this);
	this->updaterAction = new QAction(tr("Updater"), this);
	this->neighboursAction = new QAction(tr("Neighbours"), this);
	this->aboutAction = new QAction(tr("About"), this);
	this->quitAction = new QAction(tr("Quit"), this);
	QObject::connect(this->quitAction, SIGNAL(triggered()), this, SLOT(store_settings_and_quit()));
	QObject::connect(this->notificationAction, SIGNAL(triggered()), this, SLOT(activateNotificationTab()));
	QObject::connect(this->statisticsAction, SIGNAL(triggered()), this, SLOT(activateStatisticsTab()));
	QObject::connect(this->deviceAction, SIGNAL(triggered()), this, SLOT(activateDeviceTab()));
	QObject::connect(this->updaterAction, SIGNAL(triggered()), this, SLOT(activateUpdaterTab()));
	QObject::connect(this->neighboursAction, SIGNAL(triggered()), this, SLOT(activateNeighboursTab()));
	QObject::connect(this->settingsAction, SIGNAL(triggered()), this, SLOT(displaySettings()));
	QObject::connect(this->aboutAction, SIGNAL(triggered()), this, SLOT(displayAbout()));

	// Add items to menu
	this->menu = new QMenu(tr("Spectator"), this);
	this->menu->addAction(this->notificationAction);
	this->menu->addAction(this->statisticsAction);
	this->menu->addAction(this->deviceAction);
	this->menu->addAction(this->updaterAction);
	this->menu->addAction(this->neighboursAction);
	this->menu->addSeparator();
	this->menu->addAction(this->settingsAction);
	this->menu->addAction(this->aboutAction);
	this->menu->addSeparator();
	this->menu->addAction(this->quitAction);

	// Set the menu bar
	this->menuBar = new QMenuBar(this);
	this->menuBar->addMenu(this->menu);

	this->setMenuBar(this->menuBar);


	// Create neighbour interface menu
	this->neighbourInterfaceMenu = new QMenu;
	this->ui->neighbourInterfacesButton->setMenu(this->neighbourInterfaceMenu);
}

void MainWindow::updateSettings()
{
	this->settings->mainTabIndex = this->ui->mainTab->currentIndex();
	this->settings->backupPath = this->ui->backupLineEdit->text();
	this->settings->inactiveLimit = this->ui->neighbourInactiveSpinBox->value();
	this->settings->neighboursAdvancedView = this->ui->neighboursAdvancedViewCheckBox->isChecked();

	if (!this->dockModeActive) {
		if (this->isVisible()) {
			this->settings->geometry = this->saveGeometry();
		}
	}

	// Store preffered inteface
	QString interfaceName = this->interfaceWidget->getCurrentInterface();
	if (!interfaceName.isEmpty()) {
		this->settings->prefferedInterface = interfaceName;
	}
}

void MainWindow::useSettings(SpectatorSettings *settings)
{
	bool firstSettings = this->settings == NULL;  // just loaded after start

	this->settings = settings;
	this->ui->mainTab->setCurrentIndex(settings->mainTabIndex);
	this->ui->backupLineEdit->setText(this->settings->backupPath);
	this->ui->backupLineEdit->setToolTip(this->settings->backupPath);
	this->interfaceWidget->prefferedInterface = this->settings->prefferedInterface;
	this->ui->neighbourInactiveSpinBox->setValue(this->settings->inactiveLimit);
	this->ui->neighboursAdvancedViewCheckBox->setChecked(this->settings->neighboursAdvancedView);

	bool systrayAvailable = this->testSystrayMode();
	if (this->settings->systrayMode && !systrayAvailable) {
		log_settings.critical() << "System tray is not working properly. Unsing non-systray mode.";
	}
	bool useSystray = this->settings->systrayMode && systrayAvailable;
	this->setSystrayMode(useSystray);

	bool dockAvailable = this->testDockMode();
	if (this->settings->dock && !dockAvailable) {
		log_settings.critical() << "Dock mode is not working properly. Using non-docking mode.";
	}
	// Systray in necessary for the dock mode
	bool useDock = this->settings->dock && dockAvailable && useSystray;
	this->setDockMode(useDock);

	if (!this->isInDockMode()) {
		if (!this->settings->geometry.isEmpty())
			this->restoreGeometry(this->settings->geometry);
	}

	if (!this->isInSystrayMode()) {
		this->show();
	}

	// Run for the first time
	if (!this->settings->configPresent) {
		this->recalculateSize();
	}

	if (firstSettings && !this->settings->configPresent) {
		// no config found -> display settings dialog on startup
		this->displaySettings();
	} else {
		this->reinitConnection();
	}
}

bool MainWindow::isInSystrayMode()
{
	return this->systrayModeActive;
}

bool MainWindow::isInDockMode()
{
	return this->dockModeActive;
}

void MainWindow::showRunningInSystrayPopup()
{
	this->trayIcon->showMessage(tr("Running in system tray mode"), tr("Spectator is available in the system tray."), QSystemTrayIcon::Information, 1000 * this->settings->popupTime);
}

void MainWindow::reinitConnection()
{
	this->handleDisconneted();

	// Delete old one
	if (this->connector) {
		delete this->connector;
		this->connector = NULL;
	}

	this->connector = new Connector(this->settings->server, this->settings->port, this->settings->tokenPath);
	QObject::connect(this->connector, SIGNAL(report(Communicator::ReportType,QString)), this, SLOT(updateStatusBar(Communicator::ReportType,QString)));

	// Removed chats are deleted to avoid memory leaks
	QObject::connect(this->connector, SIGNAL(chatRemoved(Chats::Base *)), this, SLOT(handleChatRemoved(Chats::Base *)));

	// Connect connect and disconnect signals
	QObject::connect(this->connector, SIGNAL(connected()), this, SLOT(handleConneted()));
	QObject::connect(this->connector, SIGNAL(disconnected()), this, SLOT(handleDisconneted()));

	// Perform chats chats
	Chats::Initial *initialChat = new Chats::Initial();
	QObject::connect(initialChat, SIGNAL(exportCapabilities(QList<Capability>)), this, SLOT(handleCapatibilitiesLoaded(QList<Capability>)));
	this->connector->performInitialChat(initialChat);

	Chats::GetNotifications *getNotificationsChat = new Chats::GetNotifications(this->settings->language);
	QObject::connect(getNotificationsChat, SIGNAL(exportNotifications(QList<ParsedNotification>)), this, SLOT(updateNotificationList(QList<ParsedNotification>)));
	this->connector->performChat(getNotificationsChat);

	Chats::GetNethist *getNethistChat = new Chats::GetNethist();
	QObject::connect(getNethistChat, SIGNAL(exportNethist(QList<NethistRecord>)), this, SLOT(updateNethist(QList<NethistRecord>)));
	this->connector->performChat(getNethistChat);

	Chats::GetStats *getStatsChat = new Chats::GetStats();
	QObject::connect(getStatsChat, SIGNAL(exportStats(StatsRecord)), this, SLOT(updateStats(StatsRecord)));
	this->connector->performChat(getStatsChat);

	Chats::GetUpdaterData *getUpdaterChat = new Chats::GetUpdaterData(this->settings->language);
	QObject::connect(getUpdaterChat, SIGNAL(exportLists(UpdaterDataRecord)), this, SLOT(updateUpdaterData(UpdaterDataRecord)));
	this->connector->performChat(getUpdaterChat);

	Chats::GetSerial *getSerialChat = new Chats::GetSerial();
	QObject::connect(getSerialChat, SIGNAL(exportSerial(QString)), this, SLOT(updateSerial(QString)));
	this->connector->performChat(getSerialChat);

	Chats::GetNeighbours *getNeighbourChat = new Chats::GetNeighbours(this->settings->activeNeighbourDevices);
	QObject::connect(getNeighbourChat, SIGNAL(exportNeighbours(QList<NeighbourRecord>)), this, SLOT(updateNeighboursInit(QList<NeighbourRecord>)));
	this->connector->performChat(getNeighbourChat);

	// Trigger notification refresh timer
	this->refreshTimer.stop();
	if (this->settings->refreshInterval > 0)
		this->refreshTimer.start(this->settings->refreshInterval * 1000);


	this->connector->initiateConnection();
}

void MainWindow::show() {
	QMainWindow::show();
	if (this->isInDockMode()) {
		this->recalculateSize();
	}
}

void MainWindow::toggleWindow(QSystemTrayIcon::ActivationReason reason)
{
	if (reason != QSystemTrayIcon::Context) {
		if (this->isVisible()) {
			#ifdef Q_OS_WIN
			this->close();
			#else
			if (!this->isActiveWindow()) {
				this->activateWindow();
				this->raise();
			} else {
				this->close();
			}
			#endif
		} else {
			this->updateWindowPosition();
			this->activateWindow();
			this->raise();
			this->show();
			if (!this->isInDockMode()) {
				if (!this->settings->geometry.isEmpty())
					this->restoreGeometry(this->settings->geometry);
			}
		}
	}
}

void MainWindow::handleChatRemoved(Chats::Base *chat)
{
	delete chat;
}

void MainWindow::closeEvent(QCloseEvent *event)
{
	this->updateSettings();
	if (!this->isInSystrayMode()) {
		this->settings->store();
	}
	event->accept();
}

void MainWindow::store_settings_and_quit()
{
	this->updateSettings();
	this->settings->store();
	qApp->quit();
}

void MainWindow::activateNotificationTab()
{
	this->updateWindowPosition();
	this->show();
	this->ui->mainTab->setCurrentIndex(TAB_IDX_NOTIFICATION);
}

void MainWindow::activateStatisticsTab()
{
	this->updateWindowPosition();
	this->show();
	this->ui->mainTab->setCurrentIndex(TAB_IDX_NETHIST);
}

void MainWindow::activateDeviceTab()
{
	this->updateWindowPosition();
	this->show();
	this->ui->mainTab->setCurrentIndex(TAB_IDX_STATS);
}

void MainWindow::activateUpdaterTab()
{
	this->updateWindowPosition();
	this->show();
	this->ui->mainTab->setCurrentIndex(TAB_IDX_UPDATER);
}

void MainWindow::activateNeighboursTab()
{
	this->updateWindowPosition();
	this->show();
	this->ui->mainTab->setCurrentIndex(TAB_IDX_NEIGHBOURS);
}

void MainWindow::displayAbout()
{
	QMessageBox::about(
		this, tr("About Spectator"), tr(
			"<h1>Spectator</h1>"
			"Version: ") + SPECTATOR_VERSION + tr(
			"<p><center>Application which communicates with the Turris router using <br/> <b>NETCONF protocol</b></center></p>"
			"<b>Uses:</b>") +
			"<ul>"
			"<li>QT " QT_VERSION_STR "</li>"
			"<li>qcustomplot 1.3.1</li>"
			"<li>OpenSSL</li>"
			"</ul>"
			"<center>Copyright © 2015 <a href='https://www.nic.cz'>CZ.NIC, z. s. p. o.</a></center>"
			"<br/>"
	);
}

void MainWindow::displayBackupBrowseDialog()
{
	QString filename = QFileInfo(this->ui->backupLineEdit->text()).absoluteFilePath();
	QString res = QFileDialog::getSaveFileName(
		this, tr("backup file path"), filename, tr("Backups (*.tar.bz2);;All files (*.*)"),
		0, QFileDialog::DontConfirmOverwrite
	);
	if (!res.isNull()) {
		this->ui->backupLineEdit->setText(res);
		this->ui->backupLineEdit->setToolTip(res);
	}
}

void MainWindow::displaySettings()
{
	SettingsDialog *settingsDialog = new SettingsDialog();
	this->updateSettings();
	settingsDialog->loadSettings(*this->settings);

	// Don't try to reconnect when the settings are displayed
	this->reconnectTimer.stop();

	// Disable neighbour editing
	this->callNeighbourMacEditDisable();

	// Stop refresh when the settings is displayed
	this->refreshTimer.stop();
	if (settingsDialog->exec()) {
		QString origLang = this->settings->language;
		bool systrayModeBefore = this->isInSystrayMode();
		log_settings.debug() << "Applying settings from the settings dialog.";
		settingsDialog->updateSettings(*this->settings);
		this->settings->store();
		this->useSettings(this->settings);
		if (this->isInSystrayMode() && !systrayModeBefore) {
			this->showRunningInSystrayPopup();
		}
		if (origLang != this->settings->language) {
			emit this->translationNeeded();
		}
	} else {
		this->refreshTimer.start(RECONNECT_TIMEOUT);
		if (this->settings->refreshInterval > 0)
			this->refreshTimer.start(this->settings->refreshInterval * 1000);
	}

	delete settingsDialog;
}

void MainWindow::updateStatusBar(Communicator::ReportType type, QString msg)
{
	this->ui->statusTextLabel->setText(msg);
	if (Communicator::RT_ERROR == type) {
		this->ui->statusIconLabel->setPixmap(QApplication::style()->standardIcon(QStyle::SP_MessageBoxCritical).pixmap(25,25));
	} else if (Communicator::RT_WARNING == type) {
		this->ui->statusIconLabel->setPixmap(QApplication::style()->standardIcon(QStyle::SP_MessageBoxWarning).pixmap(25,25));
	} else {
		this->ui->statusIconLabel->setPixmap(QApplication::style()->standardIcon(QStyle::SP_MessageBoxInformation).pixmap(25,25));
	}
}

void MainWindow::updateStatusBar(QString msg) {
	this->updateStatusBar(Communicator::RT_INFO, msg);
}

void MainWindow::callPerformLoopChats()
{
	if (!this->connector || !this->connector->ready()) {
		return;
	}

	// Always refresh notifications
	Chats::GetNotifications *getNotificationsChat = new Chats::GetNotifications(this->settings->language);
	QObject::connect(getNotificationsChat, SIGNAL(exportNotifications(QList<ParsedNotification>)), this, SLOT(updateNotificationList(QList<ParsedNotification>)));
	this->connector->performChat(getNotificationsChat);

	// Always refresh neighbours
	this->callNeighbourChat();

	// Refresh other only it concerns selected tab
	switch (this->ui->mainTab->currentIndex()) {
		case TAB_IDX_NETHIST:
			{
				Chats::GetNethist *getNethistChat = new Chats::GetNethist();
				QObject::connect(getNethistChat, SIGNAL(exportNethist(QList<NethistRecord>)), this, SLOT(updateNethist(QList<NethistRecord>)));
				this->connector->performChat(getNethistChat);
			}
			break;

		case TAB_IDX_STATS:
			{
				Chats::GetStats *getStatsChat = new Chats::GetStats();
				QObject::connect(getStatsChat, SIGNAL(exportStats(StatsRecord)), this, SLOT(updateStats(StatsRecord)));
				this->connector->performChat(getStatsChat);
			}
			break;

		case TAB_IDX_UPDATER:
			{
				if (!this->updaterUserList->isInEditMode()) {
					Chats::GetUpdaterData *getUpdaterChat = new Chats::GetUpdaterData(this->settings->language);
					QObject::connect(getUpdaterChat, SIGNAL(exportLists(UpdaterDataRecord)), this, SLOT(updateUpdaterData(UpdaterDataRecord)));
					this->connector->performChat(getUpdaterChat);
				}
			}
		break;
	}
}

void MainWindow::callNeighbourChat() {
	if (this->connector && this->connector->ready() && this->enableNeighbourChat) {
		Chats::GetNeighbours *getNeighbourChat = new Chats::GetNeighbours(this->settings->activeNeighbourDevices);
		QObject::connect(getNeighbourChat, SIGNAL(exportNeighbours(QList<NeighbourRecord>)), this, SLOT(updateNeighbours(QList<NeighbourRecord>)));
		this->connector->performChat(getNeighbourChat);
	}
}

void MainWindow::handleRebootingRouter()
{
	log_control.debug() << "Router started to reboot.";
	this->setNotificationsEnabled(false);
	this->ui->controlScrollArea->setEnabled(false);
}

void MainWindow::updateWindowPosition()
{
	if (!this->isInDockMode()) {  // update only when in dock mode
		return;
	}

	QSize desktopSize = QApplication::desktop()->size();
	QRect trayGemoetry = this->trayIcon->geometry();
	int x1, x2, y1, y2;
	trayGemoetry.getCoords(&x1, &y1, &x2, &y2);
	int window_width = this->width();
	int window_height = this->height();

	// left vs right
	bool screen_left;
	if (desktopSize.width() - x1 > x1) {
		screen_left = true;
	} else {
		screen_left = false;
	}

	// top vs bottom
	bool screen_top;
	if (desktopSize.height() - y1 > y1) {
		screen_top = true;
	} else {
		screen_top = false;
	}

	// horizontal vs vertical
	bool horizontal;
	if (screen_top) {
		if (screen_left) {  // top left
			horizontal = x1 > y1;
		} else { // top right
			horizontal = desktopSize.width() - x2 > y1;
		}
	} else {
		if (screen_left) { // bottom left
			horizontal = x1 > desktopSize.height() - y1;
		} else { // bottom right
			horizontal = desktopSize.width() - x2 > desktopSize.height() - y2;
		}
	}

	// set the coordinates
	int window_x, window_y;
	if (screen_top) {
		if (screen_left) {  // top left
			window_x = horizontal ? x1 : x2;
			window_y = horizontal ? y2 : y1;
		} else { // top right
			window_x = horizontal ? x2 - window_width : x1 - window_width;
			window_y = horizontal ? y2 : y1;
		}
	} else {
		if (screen_left) { // bottom left
			window_x = horizontal ? x1 : x2;
			window_y = horizontal ? y1 - window_height : y2 - window_height;
		} else { // bottom right
			window_x = horizontal ? x2 - window_width : x1 - window_width;
			window_y = horizontal ? y1 - window_height : y2 - window_height;
		}
	}

	this->setGeometry(window_x, window_y, this->width(), this->height());
}

void MainWindow::updateNotificationList(QList<ParsedNotification> list)
{
	// hide/show noNotification label
	this->setNoNotificationsVisible(list.empty());

	QList<NotificationWidget *> widgets = this->ui->notificationScrollAreaWidgetContents->findChildren<NotificationWidget *>();
	log_notifications.debug() << "Updating notifications";
	for (QList<ParsedNotification>::iterator i = list.begin(); i != list.end(); ++i) {
		log_notifications.debug() << "Downloaded notification:" << i->id << i->severity;
	}

	// Add new
	int newCount = 0;
	for (QList<ParsedNotification>::iterator j = list.begin(); j != list.end(); ++j) {
		bool inList = false;
		for (QList<NotificationWidget *>::iterator i = widgets.begin(); i != widgets.end(); ++i) {
			if (j->id == (*i)->id) {
				inList = true;
				(*i)->updateTexts((*j).body);
			}
		}
		if (!inList) {
			new NotificationWidget(this->ui->notificationScrollAreaWidgetContents, this, j->id, j->severity, j->time, j->body);
			newCount += 1;
		}
	}

	// Remove missing
	for (QList<NotificationWidget *>::iterator i = widgets.begin(); i != widgets.end(); ++i) {
		bool inList = false;
		for (QList<ParsedNotification>::iterator j = list.begin(); j != list.end(); ++j) {
			if (j->id == (*i)->id)
				inList = true;
		}
		if (!inList)
			delete(*i);
	}

	if (newCount && this->isInSystrayMode()) {
		this->trayIcon->showMessage(tr("Notifications lists updated"), tr("Number of new notifications: ") + QString::number(newCount), QSystemTrayIcon::Information, 1000 * this->settings->popupTime);
		this->trayIcon->disconnect(SIGNAL(messageClicked()));
		QObject::connect(this->trayIcon, SIGNAL(messageClicked()), this, SLOT(handleSystrayNotificationMessageClicked()));
	}
	log_notifications.debug() << "Notifications updated.";
}

void MainWindow::updateNethist(QList<NethistRecord> list)
{
	log_nethist.debug() << "Updating nethist";

	if (list.empty()) {  // No data returned..
		log_nethist.debug() << "no data returned";
		return;
	}

	NethistRecord last = list.last();
	this->ui->loadText->setText(QString("%1").arg(last.cpuLoad));
	this->ui->cpuTemperatureText->setText(QString::number(last.cpuTemperature));
	this->ui->boardTemperatureText->setText(QString::number(last.boardTemperature));
	this->ui->usedDiskText->setText(QString::number(last.usedDiskSpace / 1024) + " MB");  // in MB
	this->ui->freeDiskText->setText(QString::number(last.availableDiskSpace / 1024) + " MB");  // in MB
	this->ui->totalDiskText->setText(QString::number((last.availableDiskSpace + last.usedDiskSpace) / 1024) + " MB");  // in MB
	this->ui->memoryTotalText->setText(QString::number(last.totalMemory / 1024) + " MB");  // in MB
	this->ui->memoryFreeText->setText(QString::number(last.freeMemory / 1024) + " MB");  // in MB
	this->ui->memoryUsedText->setText(QString::number((last.totalMemory - last.freeMemory - last.cachedMemory) / 1024) + " MB");  // in MB
	this->ui->memoryCachedText->setText(QString::number(last.cachedMemory / 1024) + " MB");  // in MB

	QVector<double> xAxis;
	QVector<double> yDiskTotalAxis;
	QVector<double> yDiskUsedAxis;
	QVector<double> yLoadAxis;
	QVector<double> yTemperatureBoardAxis;
	QVector<double> yTemperatureCpuAxis;
	QVector<double> yMemoryTotalAxis;
	QVector<double> yMemoryUsedAxis;
	QVector<double> yMemoryCachedAxis;
	QMap<QString, InterfaceWidgetRecord> interfaceData;

	QMap<QString, qulonglong> lastRx, lastTx, lastTime;
	foreach (NethistRecord record, list) {
		xAxis.append(record.time.toTime_t());
		yDiskUsedAxis.append(double(record.usedDiskSpace) / 1024);
		yDiskTotalAxis.append(double(record.usedDiskSpace + record.availableDiskSpace) / 1024);
		yLoadAxis.append(record.cpuLoad);
		yTemperatureCpuAxis.append(record.cpuTemperature);
		yTemperatureBoardAxis.append(record.boardTemperature);
		yMemoryTotalAxis.append(record.totalMemory / 1024);
		yMemoryUsedAxis.append((record.totalMemory - record.freeMemory - record.cachedMemory) / 1024);
		yMemoryCachedAxis.append((record.totalMemory - record.freeMemory) / 1024);

		foreach (NethistInterfaceRecord interfaceRecord, record.interfaces) {
			if (!lastTime.contains(interfaceRecord.name)) {  // skip first
				lastRx[interfaceRecord.name] = interfaceRecord.rx;
				lastTx[interfaceRecord.name] = interfaceRecord.tx;
				lastTime[interfaceRecord.name] = record.time.toTime_t();
				continue;
			}
			if (!interfaceData.contains(interfaceRecord.name)) {
				interfaceData[interfaceRecord.name] = InterfaceWidgetRecord();
			}
			int time = record.time.toTime_t();
			InterfaceWidgetRecord * iRecord = &interfaceData[interfaceRecord.name];
			iRecord->xTimeAxis.append(time);
			iRecord->yRxAxis.append(
				(interfaceRecord.rx - lastRx[interfaceRecord.name])
				/ (time - lastTime[interfaceRecord.name])
			);
			iRecord->yTxAxis.append(
				(interfaceRecord.tx - lastTx[interfaceRecord.name])
				/ (time - lastTime[interfaceRecord.name])
			);
			lastRx[interfaceRecord.name] = interfaceRecord.rx;
			lastTx[interfaceRecord.name] = interfaceRecord.tx;
			lastTime[interfaceRecord.name] = time;
		}
	}

	log_nethist.debug() << "Nethist records parsed: " << list.count();

	this->interfaceWidget->setData(interfaceData);

	QCustomPlot *plot;
	plot = this->ui->diskPlot;
	{
		int maxDisk = 0;
		foreach (float load, yDiskTotalAxis) {
			maxDisk = load > maxDisk ? load : maxDisk;
		}
		plot->axisRect()->axis(QCPAxis::atLeft)->setRange(0, maxDisk * 1.1);
		plot->axisRect()->axis(QCPAxis::atBottom)->setRange(xAxis.first(), xAxis.last());
		plot->graph(0)->setData(xAxis, yDiskTotalAxis);
		plot->graph(1)->setData(xAxis, yDiskUsedAxis);
		plot->replot();
	}

	plot = this->ui->loadPlot;
	{
		double maxLoad = 0;
		foreach (float load, yLoadAxis) {
			maxLoad = load > maxLoad ? load : maxLoad;
		}
		plot->axisRect()->axis(QCPAxis::atLeft)->setRange(0, maxLoad * 1.1);
		plot->axisRect()->axis(QCPAxis::atBottom)->setRange(xAxis.first(), xAxis.last());
		plot->graph(0)->setData(xAxis, yLoadAxis);
		plot->replot();
	}

	plot = this->ui->temperaturePlot;
	{
		int maxTemp = 0;
		int minTemp = 150;
		foreach (int temp, yTemperatureCpuAxis) {
			maxTemp = temp > maxTemp ? temp : maxTemp;
			minTemp = temp < minTemp ? temp : minTemp;
		}
		foreach (int temp, yTemperatureBoardAxis) {
			maxTemp = temp > maxTemp ? temp : maxTemp;
			minTemp = temp < minTemp ? temp : minTemp;
		}
		plot->axisRect()->axis(QCPAxis::atLeft)->setRange(minTemp * 0.8, maxTemp * 1.1);
		plot->axisRect()->axis(QCPAxis::atBottom)->setRange(xAxis.first(), xAxis.last());
		plot->graph(0)->setData(xAxis, yTemperatureCpuAxis);
		plot->graph(1)->setData(xAxis, yTemperatureBoardAxis);
		plot->replot();
	}

	plot = this->ui->memoryPlot;
	{
		double maxMem = 0;
		foreach (float mem, yMemoryTotalAxis) {
			maxMem = mem > maxMem ? mem : maxMem;
		}
		plot->axisRect()->axis(QCPAxis::atLeft)->setRange(0, maxMem * 1.1);
		plot->axisRect()->axis(QCPAxis::atBottom)->setRange(xAxis.first(), xAxis.last());
		plot->graph(0)->setData(xAxis, yMemoryTotalAxis);
		plot->graph(1)->setData(xAxis, yMemoryCachedAxis);
		plot->graph(2)->setData(xAxis, yMemoryUsedAxis);
		plot->replot();
	}

	log_nethist.debug() << "Charts updated";
}


void MainWindow::updateStats(StatsRecord record)
{
	// TODO move to a better place
	QHash<QString, QString> typeAliases;
	typeAliases["inet6"] = "IPv6";
	typeAliases["inet"] = "IPv4";
	typeAliases["link/ether"] = "MAC";

	log_stats.debug() << "Updating stats.";
	setTextLabel(this->ui->kernelLabel, record.kernelVersion);
	setTextLabel(this->ui->versionLabel, record.osVersion);
	setTextLabel(this->ui->typeLabel, record.model + " - " + record.boardName.toUpper());
	setTextLabel(this->ui->hostnameLabel, record.hostname);
	int days = record.uptime / 24 / 60 / 60;
	int hours = (record.uptime - 24 * 60 * 60 * days) / 60 / 60;
	int minutes = (record.uptime - 24 * 60 * 60  * days - hours * 60 * 60) / 60;
	setTextLabel(this->ui->uptimeLabel, QString("%1d %2h %3m").arg(QString::number(days)).arg(QString::number(hours)).arg(QString::number(minutes)));
	setTextLabel(this->ui->firewallSendingLabel, QString(tr("%1 (%2 second ago)")).arg(record.firewallStatus).arg(record.firewallAge));
	setTextLabel(this->ui->ucollectSendingLabel, QString(tr("%1 (%2 second ago)")).arg(record.ucollectStatus).arg(record.ucollectAge));

	// remove all widgets
	foreach (QWidget *w, this->ui->networkDevicesList->findChildren<QWidget *>()) {
		delete w;
	}

	// remove the spacer
	QSpacerItem *spacer = this->ui->networkDevicesList->layout()->itemAt(0)->spacerItem();
	this->ui->networkDevicesList->layout()->removeItem(spacer);
	delete spacer;

	log_stats.debug() << "Interface count:" << record.interfaces.count();
	QList<QString> interfaceNames;
	int active_count = 0;
	for (QList<StatsInterface>::const_iterator i = record.interfaces.begin(); i != record.interfaces.end(); ++i) {
		interfaceNames.append(i->name);
		// Display only device which are up
		if (!i->up) {
			continue;
		}
		active_count +=1;

		QLabel *l = new QLabel(this->ui->networkDevicesList);
		l->setTextInteractionFlags(Qt::TextSelectableByMouse | Qt::LinksAccessibleByMouse);
		QString text = "<b>" + i->name + "</b>";
		text += "<table>";
		for (QList<StatsAddress>::const_iterator j = i->addresses.begin(); j != i->addresses.end(); ++j) {
			QString typeAlias = typeAliases[j->type];
			text += "<tr>";
			text += "<td>" + (typeAlias.isEmpty() ? j->type : typeAlias) + ":</td>";
			text += "<td><i>" + j->address + "</i></td>";
			text += "</tr>";
		}

		text += "</table>";
		l->setText(text);
		this->ui->networkDevicesList->layout()->addWidget(l);
	}
	this->ui->networkDevicesList->layout()->addItem(new QSpacerItem(40, 20, QSizePolicy::Minimum, QSizePolicy::Expanding));
	log_stats.debug() << "Active Interface count:" << active_count;

	log_stats.debug() << "Updating interface list.";
	this->updateNeighbourDevices(interfaceNames);

	log_stats.debug() << "Stats were updated.";
}

void MainWindow::updateUpdaterData(UpdaterDataRecord record)
{
	this->ui->updaterStatusLabel->setText(record.status);
	this->updaterUserList->setUserListData(record.userLists);
	this->updaterActivities->setActivitiesData(record.lastActivities);
}

void MainWindow::updateSerial(QString serial)
{
	bool res;
	serial = (serial.length() % 2 == 0 ? "0x" : "0x0") + serial;
	qlonglong serialDec = serial.toULongLong(&res, 16);
	if (res) {
		this->ui->serialLabel->setText(QString::number(serialDec));
	}
}

void MainWindow::updateNeighbours(QList<NeighbourRecord> list, bool popup, bool markUnread)
{
	// Log downloaded neighbours
	log_neighbours.debug() << "Updating neighbours";
	log_neighbours.debug() << "Interface filter used:" << this->settings->activeNeighbourDevices;
	foreach (NeighbourRecord record, list) {
		log_neighbours.debug() << "Downloaded neighbour:" << record.macAddress << record.interfaceName;
	}

	// filter list based on time limit
	QList<NeighbourRecord> filteredList;
	foreach (NeighbourRecord record, list) {
		NeighbourRecord newRecord(record);
		newRecord.ips.clear();
		foreach (NeighbourIpRecord ipRecord, record.ips) {
			// At leas one of the Since times is smaller then limit
			// Since times are -1 when statistics tag is not present
			int limit = this->ui->neighbourInactiveSpinBox->value();
			if (ipRecord.updatedSince < limit || ipRecord.usedSince < limit || ipRecord.confirmedSince < limit) {
				newRecord.ips.append(ipRecord);
			} else {
				log_neighbours.debug() << "Record is outdated and will be skipped" << record.macAddress << record.interfaceName << ipRecord.ipAddress;
			}
		}

		if (!newRecord.ips.isEmpty())
			filteredList.append(newRecord);
	}

	// hide/show noNeighbours label
	this->setNoNeighboursVisible(filteredList.empty());

	QList<NeighbourWidget *> widgets = this->ui->neighboursScrollWidget->findChildren<NeighbourWidget *>();

	// Add new and update existing
	QStringList addedDevices;
	foreach (NeighbourRecord record, filteredList) {
		bool inList = false;
		foreach (NeighbourWidget *widget, widgets) {
			if (widget->mac == record.macAddress && widget->dev == record.interfaceName) {
				// Update name mapping
				widget->setMacMapping(&this->settings->neighboursMacToName);
				// Updating record
				widget->setContent(&record);
				widget->setBodyText(this->settings->neighboursAdvancedView);
				QObject::connect(widget, SIGNAL(editStatusChanged(bool)), this, SLOT(handleNeighbourMacEdit(bool)));
				inList = true;
			}
		}
		if (!inList) {
			NeighbourWidget *widget = new NeighbourWidget(this->ui->neighboursScrollWidget,
					this, &record, &this->settings->neighboursMacToName);
			widget->setBodyText(this->settings->neighboursAdvancedView);
			((QVBoxLayout *) this->ui->neighboursScrollWidget->layout())->insertWidget(0, widget);
			if (markUnread) {
				widget->markUnread();
			} else {
				widget->markRead();
			}
			QString alias = widget->getAlias(widget->mac);
			addedDevices.append(alias.isEmpty() ? widget->mac : alias);
		}
	}

	// Remove missing
	QStringList removedDevices;
	foreach (NeighbourWidget *widget, widgets) {
		bool inList = false;
		foreach (NeighbourRecord record, filteredList) {
			if (widget->mac == record.macAddress && widget->dev == record.interfaceName) {
				inList = true;
			}
		}
		if (!inList) {
			QString alias = widget->getAlias(widget->mac);
			removedDevices.append(alias.isEmpty() ? widget->mac : alias);
			delete widget;
		}
	}

	if (this->isInSystrayMode() && popup && (removedDevices.count()|| addedDevices.count())) {
		QString msg;
		msg += tr("discovered: ") + QString::number(addedDevices.count()) + "\n";
		foreach (QString device, addedDevices) {
			msg += " • " + device + "\n";
		}
		msg += tr("disappeared: ") + QString::number(removedDevices.count())  + "\n";
		foreach (QString device, removedDevices) {
			msg += " • " + device + "\n";
		}

		this->trayIcon->showMessage(
			tr("Neighbour list has changed."), msg, QSystemTrayIcon::Information, 1000 * this->settings->popupTime);
		this->trayIcon->disconnect(SIGNAL(messageClicked()));
		QObject::connect(this->trayIcon, SIGNAL(messageClicked()), this, SLOT(handleSystrayNeighbourClicked()));
	}

	log_neighbours.debug() << "Neighbours updated.";
}

void MainWindow::updateNeighboursInit(QList<NeighbourRecord> list)
{
	this->updateNeighbours(list, false, false);
}

void MainWindow::setNoNotificationsVisible(bool visible)
{
	this->ui->notificationScrollAreaWidgetContents->setSizePolicy(QSizePolicy::Expanding, visible ? QSizePolicy::Expanding : QSizePolicy::Fixed);
	this->ui->noNotificationsLabel->setVisible(visible);
}

void MainWindow::setNoNeighboursVisible(bool visible)
{
	this->ui->neighboursScrollWidget->setSizePolicy(QSizePolicy::Expanding, visible ? QSizePolicy::Expanding : QSizePolicy::Fixed);
	this->ui->noNeighboursLabel->setVisible(visible);
}

bool MainWindow::testDockMode()
{
	QRect rect = this->trayIcon->geometry();
	return rect.width() > 0 && rect.height() > 0;
}

bool MainWindow::testSystrayMode()
{
	if (this->settings) {
		if (this->settings->noSystrayMode) {
			return false;
		} else if (this->settings->forceSystrayMode) {
			return true;
		}
	}
	return QSystemTrayIcon::isSystemTrayAvailable();
}

void MainWindow::setSystrayMode(bool active)
{
	if (active) {

		this->trayIcon->show();

		// Don't close on exit
		qApp->setQuitOnLastWindowClosed(false);


		this->systrayModeActive = true;
	} else {

		this->trayIcon->hide();

		// Close on exit
		qApp->setQuitOnLastWindowClosed(true);
		this->show();

		this->systrayModeActive = false;
	}
}

void MainWindow::setDockMode(bool active)
{
	unsigned int framelessFlags = Qt::FramelessWindowHint | Qt::MSWindowsFixedSizeDialogHint | Qt::CustomizeWindowHint;

	if (active) {
		// Don't switch window params when it's not necessary
		if (!((this->windowFlags() & framelessFlags) == framelessFlags)) {

			// Remove window decorations
			this->setWindowFlags(Qt::MSWindowsFixedSizeDialogHint);
			this->setWindowFlags(Qt::CustomizeWindowHint);
			this->setWindowFlags(Qt::FramelessWindowHint);

			// Set status bar icon margins
			this->ui->statusIconLayout->setContentsMargins(5, 3, 0, 5);

			this->menuBar->hide();
		}

		this->dockModeActive = true;
		// update the main window position
		this->updateWindowPosition();
	} else {
		// Don't switch window params when it's not necessary
		if (this->windowFlags() & framelessFlags) {
			// Use window decorations
			this->setWindowFlags(this->windowFlags() & ~Qt::MSWindowsFixedSizeDialogHint);
			this->setWindowFlags(this->windowFlags() & ~Qt::CustomizeWindowHint);
			this->setWindowFlags(this->windowFlags() & ~Qt::FramelessWindowHint);

			// Remove status bar icon margins
			this->ui->statusIconLayout->setContentsMargins(0, 0, 0, 0);

			this->menuBar->show();
		}
		this->dockModeActive = false;
	}
}

void MainWindow::initCharts()
{

	QCustomPlot *plot;


	// Disk plot
	plot = this->ui->diskPlot;
	{
		plot->plotLayout()->clear();
		plot->plotLayout()->setMargins(QMargins(0,0,0,0));
		QCPAxisRect *wideAxisRect = new QCPAxisRect(plot, false);
		wideAxisRect->setMargins(QMargins(0,0,0,0));
		QCPAxis *leftAxis = wideAxisRect->addAxis(QCPAxis::atLeft);
		leftAxis->setPadding(0);
		leftAxis->setRange(0.0, 249.0);
		QCPAxis *bottomAxis = wideAxisRect->addAxis(QCPAxis::atBottom);
		bottomAxis->setVisible(false);
		plot->plotLayout()->addElement(0, 0, wideAxisRect);
		QCPGraph *graph;
		// Add total
		graph = plot->addGraph(plot->axisRect()->axis(QCPAxis::atBottom), plot->axisRect()->axis(QCPAxis::atLeft));
		graph->setPen(QPen(Qt::green));
		graph->setBrush(QBrush(QColor(0x80, 0XFF, 0x80, 0x80)));
		// Add used
		graph = plot->addGraph(plot->axisRect()->axis(QCPAxis::atBottom), plot->axisRect()->axis(QCPAxis::atLeft));
		graph->setPen(QPen(Qt::red));
		graph->setBrush(QBrush(QColor(0XFF, 0x80, 0x80, 0x80)));

		plot->graph(0)->setChannelFillGraph(plot->graph(1));
	}

	// Load plot
	plot = this->ui->loadPlot;
	{
		plot->plotLayout()->clear();
		plot->plotLayout()->setMargins(QMargins(0,0,0,0));
		QCPAxisRect *wideAxisRect = new QCPAxisRect(plot, false);
		wideAxisRect->setMargins(QMargins(0,0,0,0));
		QCPAxis *leftAxis = wideAxisRect->addAxis(QCPAxis::atLeft);
		leftAxis->setPadding(0);
		leftAxis->setRange(0.0, 1.0);
		leftAxis->setAutoTickCount(3);
		QCPAxis *bottomAxis = wideAxisRect->addAxis(QCPAxis::atBottom);
		bottomAxis->setVisible(false);
		plot->plotLayout()->addElement(0, 0, wideAxisRect);
		QCPGraph *graph;
		// Add load
		graph = plot->addGraph(plot->axisRect()->axis(QCPAxis::atBottom), plot->axisRect()->axis(QCPAxis::atLeft));
		graph->setPen(QPen(Qt::black));
		graph->setBrush(QBrush(QColor(0xD0, 0xD0, 0xD4, 0x80)));
	}

	// Temperature plot
	plot = this->ui->temperaturePlot;
	{
		plot->plotLayout()->clear();
		plot->plotLayout()->setMargins(QMargins(0,0,0,0));
		QCPAxisRect *wideAxisRect = new QCPAxisRect(plot, false);
		wideAxisRect->setMargins(QMargins(0,0,0,0));
		QCPAxis *leftAxis = wideAxisRect->addAxis(QCPAxis::atLeft);
		leftAxis->setPadding(0);
		leftAxis->setRange(0.0, 150);
		leftAxis->setAutoTickCount(3);
		QCPAxis *bottomAxis = wideAxisRect->addAxis(QCPAxis::atBottom);
		bottomAxis->setVisible(false);
		plot->plotLayout()->addElement(0, 0, wideAxisRect);
		QCPGraph *graph;
		// Add board temperature
		graph = plot->addGraph(plot->axisRect()->axis(QCPAxis::atBottom), plot->axisRect()->axis(QCPAxis::atLeft));
		graph->setPen(QPen(Qt::red));
		// Add board temperature
		graph = plot->addGraph(plot->axisRect()->axis(QCPAxis::atBottom), plot->axisRect()->axis(QCPAxis::atLeft));
		graph->setPen(QPen(Qt::blue));
	}

	// Memory plot
	plot = this->ui->memoryPlot;
	{
		plot->plotLayout()->clear();
		plot->plotLayout()->setMargins(QMargins(0,0,0,0));
		QCPAxisRect *wideAxisRect = new QCPAxisRect(plot, false);
		wideAxisRect->setMargins(QMargins(0,0,0,0));
		QCPAxis *leftAxis = wideAxisRect->addAxis(QCPAxis::atLeft);
		leftAxis->setPadding(0);
		leftAxis->setRange(0.0, 2048);
		QCPAxis *bottomAxis = wideAxisRect->addAxis(QCPAxis::atBottom);
		bottomAxis->setVisible(false);
		leftAxis->setAutoTickCount(5);
		plot->plotLayout()->addElement(0, 0, wideAxisRect);
		QCPGraph *graph;
		// Add Free memory
		graph = plot->addGraph(plot->axisRect()->axis(QCPAxis::atBottom), plot->axisRect()->axis(QCPAxis::atLeft));
		graph->setPen(QPen(Qt::green));
		graph->setBrush(QBrush(QColor(0x80, 0XFF, 0x80, 0x80)));
		// Add cached memory
		graph = plot->addGraph(plot->axisRect()->axis(QCPAxis::atBottom), plot->axisRect()->axis(QCPAxis::atLeft));
		graph->setPen(QPen(Qt::gray));
		graph->setBrush(QBrush(QColor(0xD0, 0xD0, 0xD4, 0x80)));
		// Add used memory
		graph = plot->addGraph(plot->axisRect()->axis(QCPAxis::atBottom), plot->axisRect()->axis(QCPAxis::atLeft));
		graph->setPen(QPen(Qt::red));
		graph->setBrush(QBrush(QColor(0XFF, 0x80, 0x80, 0x80)));

		plot->graph(0)->setChannelFillGraph(plot->graph(1));
		plot->graph(1)->setChannelFillGraph(plot->graph(2));
	}
}

void MainWindow::discardNotification(QString messageId)
{
	QList<NotificationWidget *> widgets = this->ui->notificationScrollAreaWidgetContents->findChildren<NotificationWidget *>();
	for (QList<NotificationWidget *>::iterator i = widgets.begin(); i != widgets.end(); ++i) {
		if ((*i)->id == messageId) {
			delete *i;
			if (widgets.length() == 1)  // last notification removed
				this->setNoNotificationsVisible(true);
			break;
		}
	}
}

void MainWindow::setNotificationsEnabled(bool enabled)
{
	QList<NotificationWidget *> widgets = this->ui->notificationScrollAreaWidgetContents->findChildren<NotificationWidget *>();
	for (QList<NotificationWidget *>::iterator i = widgets.begin(); i != widgets.end(); ++i) {
		((NotificationWidget *)(*i))->setEnabledChildren(enabled);
	}
	this->ui->notificationsScrollArea->setEnabled(enabled);
}

void MainWindow::callCloseNotification(QString messageId)
{
	Chats::DeleteNotification *chat = new Chats::DeleteNotification(messageId);
	QObject::connect(chat, SIGNAL(notificationDeleted(QString)), this, SLOT(discardNotification(QString)));
	this->connector->performChat(chat);
}

void MainWindow::callRebootRouter()
{
	if (QMessageBox::Yes == QMessageBox::question(
			this, tr("Reboot the router?"),
			tr("Do you really want to reboot the router? Note that all current notifications will be discarded afterwards."),
			QMessageBox::Yes|QMessageBox::No)) {
		Chats::RebootRouter *chat = new Chats::RebootRouter();
		QObject::connect(chat, SIGNAL(rebootingRouter()), this, SLOT(handleRebootingRouter()));
		log_control.debug() << "Reboot command issued.";
		this->connector->performChat(chat);
	}
}

void MainWindow::callCheckUpdater() {
	Chats::CheckUpdater *checkChat = new Chats::CheckUpdater();
	QObject::connect(checkChat, SIGNAL(finished(bool)), this, SLOT(updaterCheckFinished(bool)));
	Chats::GetUpdaterData *getChat = new Chats::GetUpdaterData(this->settings->language);
	QObject::connect(getChat, SIGNAL(exportLists(UpdaterDataRecord)), this, SLOT(updateUpdaterData(UpdaterDataRecord)));
	this->connector->performChat(checkChat);
	this->connector->performChat(getChat);
}

void MainWindow::callSetPassword()
{
	if (this->ui->password2LineEdit->text() != this->ui->passwordLineEdit->text()) {
		QMessageBox::critical(this, tr("password mismatch"), tr("Password mismatch. Please write the same password into both fields."));
		return;
	}

	if (this->ui->password2LineEdit->text().size() < 6) {
		QMessageBox::critical(this, tr("short password"), tr("The password is too short. Please use at least <b>6 characters</b>."));
		return;
	}

	Chats::SetPassword *chat = new Chats::SetPassword(this->ui->passwordLineEdit->text());
	log_control.debug() << "Set password command started.";
	QObject::connect(chat, SIGNAL(finished(bool)), this, SLOT(setPasswordFinished(bool)));
	this->connector->performChat(chat);
}

void MainWindow::callStoreBackup()
{
	Chats::StoreBackup *chat = new Chats::StoreBackup();
	QObject::connect(chat, SIGNAL(storeData(QByteArray)), this, SLOT(storeBackupFinished(QByteArray)));
	log_control.debug() << "Store backup started.";
	this->connector->performChat(chat);
}

void MainWindow::callRestoreBackup()
{
	QFile file(this->ui->backupLineEdit->text());
	if (file.open(QFile::ReadOnly)) {
		QByteArray data = file.readAll();
		Chats::RestoreBackup *chat = new Chats::RestoreBackup(data);
		QObject::connect(chat, SIGNAL(dataRestored(QString)), this, SLOT(restoreBackupFinished(QString)));
		log_control.debug() << "Restore backup started.";
		this->connector->performChat(chat);
	} else {
		// Failed to open
		QMessageBox::critical(this, tr("failed to open"), tr("Failed to open the backup file."));
		log_control.critical() << QString("Failed to open backup file '%1'.").arg(this->ui->backupLineEdit->text());
	}
}

void MainWindow::callSetUpdater()
{
	QList<QString> newList;
	for (int i = 0; i < this->updaterUserList->getLenght(); ++i) {
		if (this->updaterUserList->getCheckbox(i)->isChecked()) {
			newList.append(*this->updaterUserList->getName(i));
		}
	}

	Chats::SetUpdater *chat = new Chats::SetUpdater(newList);
	QObject::connect(chat, SIGNAL(finished(bool)), this, SLOT(setUpdaterFinished(bool)));
	this->connector->performChat(chat);
}

void MainWindow::handleConneted()
{
	this->reconnectTimer.stop();
	this->setNotificationsEnabled(true);

	// update tray icon
	QIcon icon(":/icons/img/trayicon.png");
	this->trayIcon->setIcon(icon);
	this->ui->controlScrollArea->setEnabled(true);

	this->updaterUserList->setEnabled(true);
	this->ui->neighbourInterfacesButton->setEnabled(true);
	this->ui->neighbourInactiveSpinBox->setEnabled(true);
	this->ui->neighboursAdvancedViewCheckBox->setEnabled(true);
	this->ui->updaterStatusScrollArea->setEnabled(true);
	this->ui->neighboursScrollWidget->setEnabled(true);
	this->ui->deviceDataCollectingScrollArea->setEnabled(true);
	this->ui->deviceInterfacesScrollArea->setEnabled(true);
	this->ui->deviceSystemScrollArea->setEnabled(true);
	this->ui->statisticsMemoryFrame->setEnabled(true);
	this->ui->statisticsNandFrame->setEnabled(true);
	this->ui->statisticsNetworkFrame->setEnabled(true);
	this->ui->statisticsTemperatureFrame->setEnabled(true);

	// Call exit edit mode + call the chat
	QObject::connect(this->ui->neighbourInactiveSpinBox, SIGNAL(valueChanged(int)), this, SLOT(callNeighbourMacEditDisable()));
	QObject::connect(this->ui->neighbourInactiveSpinBox, SIGNAL(valueChanged(int)), this, SLOT(callNeighbourChat()));
}

void MainWindow::handleDisconneted()
{
	// update tray icon
	QIcon icon(":/icons/img/trayicon-gray.png");
	this->trayIcon->setIcon(icon);

	// Set timer
	this->reconnectTimer.stop();
	this->reconnectTimer.start(RECONNECT_TIMEOUT);

	this->updateStatusBar(Communicator::RT_ERROR, tr("Disconnected."));

	this->setNotificationsEnabled(false);
	this->ui->controlScrollArea->setEnabled(false);
	this->ui->neighbourInterfacesButton->setEnabled(false);
	this->ui->neighbourInactiveSpinBox->setEnabled(false);
	this->ui->neighboursAdvancedViewCheckBox->setEnabled(false);
	this->updaterUserList->setEnabled(false);
	this->ui->updaterStatusScrollArea->setEnabled(false);
	this->ui->neighboursScrollWidget->setEnabled(false);
	this->ui->deviceDataCollectingScrollArea->setEnabled(false);
	this->ui->deviceInterfacesScrollArea->setEnabled(false);
	this->ui->deviceSystemScrollArea->setEnabled(false);
	this->ui->statisticsMemoryFrame->setEnabled(false);
	this->ui->statisticsNandFrame->setEnabled(false);
	this->ui->statisticsNetworkFrame->setEnabled(false);
	this->ui->statisticsTemperatureFrame->setEnabled(false);

	// enable all tabs (capabilities of server are not known - disconnected)
	this->enableAllTabsAndMenus();

	// disconnect limit change signal
	this->ui->neighbourInactiveSpinBox->disconnect(SIGNAL(valueChanged(int)), this);
}

void MainWindow::recalculateSize()
{
	// Calculate optimal window size
	QSize sizeHint = this->sizeHint();
	sizeHint.setHeight(sizeHint.height() + 10);
	sizeHint.setWidth(sizeHint.width() + 10);
	QSize size;
	size.setHeight(this->height() > sizeHint.height() ? this->height() : sizeHint.height());
	size.setWidth(this->width() > sizeHint.width() ? this->width() : sizeHint.width());
	this->resize(size);
}

void MainWindow::handleReconnect()
{
	this->updateStatusBar(tr("Trying to reconnect."));
	this->reinitConnection();
}

void MainWindow::handleCapatibilitiesLoaded(QList<Capability> capabilities)
{
	this->updateTabsAndMenus(capabilities);
}

void MainWindow::handleSystrayNotificationMessageClicked()
{
	this->activateNotificationTab();
	this->ui->notificationsScrollArea->verticalScrollBar()->setValue(this->ui->notificationsScrollArea->verticalScrollBar()->maximum());
}

void MainWindow::handleSystrayNeighbourClicked()
{
	this->activateNeighboursTab();
	this->ui->neighboursScrollArea->verticalScrollBar()->setValue(0);
}

void MainWindow::handleNeighbourMenuItemToggled(bool checked)
{
	// get action
	QAction *action = qobject_cast<QAction *>(QObject::sender());
	QString name = action->property("interface_name").toString();
	if (this->settings->activeNeighbourDevices.contains(name)) {
		// removing
		if (!checked) {
			this->settings->activeNeighbourDevices.removeAll(name);
			log_neighbours.debug() << "Removing interface" << name << "from the filter";
		}
	} else {
		// adding
		if (checked) {
			this->settings->activeNeighbourDevices.append(name);
			log_neighbours.debug() << "Adding interface" << name << "to the filter";
		}
	}

	// remove from edit mode
	this->callNeighbourMacEditDisable();
	this->callNeighbourChat();
}

void MainWindow::handleNeighbourMacEdit(bool toEditMode)
{
	QList<NeighbourWidget *> widgets = this->ui->neighboursScrollWidget->findChildren<NeighbourWidget *>();
	if (toEditMode) {
		// disable neighbour queries
		this->enableNeighbourChat = false;
		NeighbourWidget *senderWidget = qobject_cast<NeighbourWidget *>(QObject::sender());
		foreach (NeighbourWidget *widget, widgets) {
			if (widget != senderWidget) {
				widget->setMacMapping(&this->settings->neighboursMacToName);
				widget->resetEditState();
			}
		}
	} else {
		// enable neighbour queries
		this->enableNeighbourChat = true;
		foreach (NeighbourWidget *widget, widgets) {
			widget->setMacMapping(&this->settings->neighboursMacToName);
			widget->resetEditState();
		}
	}
}

void MainWindow::handleNeighbourAdvancedViewToggled(bool advanced)
{
	this->settings->neighboursAdvancedView = advanced;
	QList<NeighbourWidget *> widgets = this->ui->neighboursScrollWidget->findChildren<NeighbourWidget *>();
	foreach (NeighbourWidget *widget, widgets) {
		widget->setBodyText(advanced);
	}
}

void MainWindow::callNeighbourMacEditDisable()
{
	this->handleNeighbourMacEdit(false);
}

void MainWindow::setPasswordFinished(bool success)
{
	if (success) {
		this->updateStatusBar(Communicator::RT_SUCCESS, tr("The password has been set successfully."));
		log_control.debug() << "Password was set.";
		this->ui->password2LineEdit->clear();
		this->ui->passwordLineEdit->clear();
	} else {
		this->updateStatusBar(Communicator::RT_ERROR, tr("Setting password failed."));
		log_control.critical() << "Failed to set password.";
	}
}

void MainWindow::updaterCheckFinished(bool success)
{
	if (success) {
		this->updateStatusBar(Communicator::RT_SUCCESS, tr("The updater check has been triggered."));
		log_updater.debug() << "Updater check was triggered.";
	} else {
		this->updateStatusBar(Communicator::RT_ERROR, tr("Failed to trigger the updater check."));
		log_updater.debug() << "Updater check was triggered.";
	}
}

void MainWindow::setUpdaterFinished(bool success)
{
	if (success) {
		this->updateStatusBar(Communicator::RT_SUCCESS, tr("The user lists of updater were successfully set."));
		log_updater.debug() << "The user lists of updater were successfully set.";
	} else {
		this->updateStatusBar(Communicator::RT_ERROR, tr("Failed to set the user lists of updater."));
		log_updater.debug() << "Failed to set the user lists of updater.";
	}
}

void MainWindow::storeBackupFinished(QByteArray data)
{
	QString path = this->ui->backupLineEdit->text();
	QFileInfo fileInfo(path);
	if (fileInfo.exists()) {
		if (fileInfo.isFile()) {
			// override dialog else return
			QMessageBox::StandardButton res = QMessageBox::warning(
				this, tr("Overwrite file"), ("File already exists. Do you want to overwrite it?"),
				QMessageBox::Ok | QMessageBox::Cancel, QMessageBox::Cancel
			);
			if (QMessageBox::Ok != res) {
				return;
			}
		} else {
			QMessageBox::critical(this, tr("file type error"), tr("'%1' is not a valid path to a file.").arg(path));
			log_control.critical() << QString("'%1' is not a file.").arg(path);
			return;
		}
	}
	QFile file(path);
	if(file.open(QFile::WriteOnly | QFile::Truncate)) {
		file.write(data);
		file.close();
		this->updateStatusBar(Communicator::RT_SUCCESS, tr("Backup stored."));
		log_control.debug() << QString("Backup was stored into '%1'.").arg(path).toStdString().c_str();
	} else {
		this->updateStatusBar(Communicator::RT_ERROR, tr("Failed to save backup file."));
		log_control.critical() << QString("Failed to save a file '%1'.").arg(path);
	}
}

void MainWindow::restoreBackupFinished(QString newIp)
{
	// update ip if present and different
	if (!newIp.isNull() && newIp != this->settings->server) {
		this->settings->server = newIp;
		// after backup restore the router will be restarted
		// program will lose connection
		// it reconnect to the new ip
		// so calling reconnect function is not needed
	}
	this->updateStatusBar(Communicator::RT_SUCCESS, tr("Backup restored."));
	log_control.debug() << QString("Backup restored from a file '%1'.").arg(this->ui->backupLineEdit->text()).toStdString().c_str();
}

void MainWindow::updateTabsAndMenus(QList<Capability> capabilities)
{
	for(QMap<int, QList<Capability> >::iterator i = this->tabCapabilities.begin(); i != this->tabCapabilities.end(); ++i) {
		foreach (Capability requiredCapability, i.value()) {
			bool present = false;
			foreach (Capability capability, capabilities) {
				if (capability.covers(requiredCapability)) {
					present = true;
					break;
				}
			}
			if (!present) {
				this->setTabAndMenusEnabled(i.key(), false);
			}
		}
	}
}

void MainWindow::enableAllTabsAndMenus()
{
	this->setTabAndMenusEnabled(TAB_IDX_NOTIFICATION, true);
	this->setTabAndMenusEnabled(TAB_IDX_NETHIST, true);
	this->setTabAndMenusEnabled(TAB_IDX_STATS, true);
	this->setTabAndMenusEnabled(TAB_IDX_UPDATER, true);
	this->setTabAndMenusEnabled(TAB_IDX_NEIGHBOURS, true);
}

void MainWindow::setTabAndMenusEnabled(int tabIndex, bool enabled)
{
	this->ui->mainTab->setTabEnabled(tabIndex, enabled);
	this->tabActions[tabIndex]->setEnabled(enabled);
}

void MainWindow::updateNeighbourDevices(QList<QString> devices)
{
	// Remove all devices and delete
	QList<QAction *> items = this->neighbourInterfaceMenu->findChildren<QAction *>();
	foreach (QAction *item, items) {
		delete item;
	}

	// Append new list
	foreach (QString device, devices) {
		QAction *action = new QAction(device, this->neighbourInterfaceMenu);
		action->setCheckable(true);
		action->setProperty("interface_name", device);
		action->setChecked(this->settings->activeNeighbourDevices.contains(device));
		QObject::connect(action, SIGNAL(toggled(bool)), this, SLOT(handleNeighbourMenuItemToggled(bool)));
		this->neighbourInterfaceMenu->addAction(action);
	}

	// Append interface without name
	QAction *action = new QAction(tr("Unknown"), this->neighbourInterfaceMenu);
	action->setCheckable(true);
	action->setProperty("interface_name", NEIGHBOURS_INTERFACE_WITHOUT_NAME);
	action->setChecked(this->settings->activeNeighbourDevices.contains(NEIGHBOURS_INTERFACE_WITHOUT_NAME));
	QObject::connect(action, SIGNAL(toggled(bool)), this, SLOT(handleNeighbourMenuItemToggled(bool)));
	this->neighbourInterfaceMenu->addAction(action);
}
