/*
 * Spectator - a desktop app which communicates with router Turris using NETCONF protocol
 * Copyright (C) 2015  CZ.NIC, z. s. p. o. <http://www.nic.cz>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef SPECTATOR_MAINWINDOW_H
#define SPECTATOR_MAINWINDOW_H

#include <QtCore>
#include <QtGui>
#include <QWidget>

#include "communicator.h"
#include "settings.h"
#include "custom_widgets.h"

namespace Ui {
	class MainWindow;
}

class MainWindow : public QMainWindow
{
	Q_OBJECT

public slots:
	void toggleWindow(QSystemTrayIcon::ActivationReason);
	void store_settings_and_quit();
	void activateNotificationTab();
	void activateStatisticsTab();
	void activateDeviceTab();
	void activateUpdaterTab();
	void activateNeighboursTab();

	void displaySettings();
	void displayAbout();
	void displayBackupBrowseDialog();

	void updateStatusBar(Communicator::ReportType type, QString msg);
	void updateStatusBar(QString msg);
	void updateNotificationList(QList<ParsedNotification> list);
	void updateNethist(QList<NethistRecord> list);
	void updateStats(StatsRecord record);
	void updateSerial(QString serial);
	void updateUpdaterData(UpdaterDataRecord record);
	void updateNeighbours(QList<NeighbourRecord> list, bool popup=true, bool markUnread=true);
	void updateNeighboursInit(QList<NeighbourRecord> list);
	void discardNotification(QString messageId);

	void callStoreBackup();
	void callRestoreBackup();
	void callCloseNotification(QString messageId);
	void callSetPassword();
	void callRebootRouter();
	void callCheckUpdater();
	void callSetUpdater();

	void setPasswordFinished(bool success);
	void updaterCheckFinished(bool soccess);
	void storeBackupFinished(QByteArray data);
	void restoreBackupFinished(QString newIp);
	void setUpdaterFinished(bool success);

public:
	SpectatorSettings *settings;
	QMap<int, QList<Capability> > tabCapabilities;
	QMenu *menu;

	void updateSettings();  // updates this->settings according to gui status
	void useSettings(SpectatorSettings *);  // load settings structure to this->settings
	bool isInSystrayMode();
	bool isInDockMode();
	void setSystrayMode(bool active);
	bool testDockMode();
	bool testSystrayMode();
	void showRunningInSystrayPopup();
	void show();
	explicit MainWindow(QWidget *parent, QSystemTrayIcon *trayIcon);
	~MainWindow();

protected:
	void contextMenuEvent(QContextMenuEvent *event) Q_DECL_OVERRIDE;
	void keyPressEvent(QKeyEvent *event);
	void closeEvent(QCloseEvent *event);
	void changeEvent(QEvent* event);

private slots:
	void handleChatRemoved(Chats::Base *chat);
	void callPerformLoopChats();
	void callNeighbourChat();
	void handleRebootingRouter();
	void handleConneted();
	void handleDisconneted();
	void handleReconnect();
	void handleSystrayNotificationMessageClicked();
	void handleSystrayNeighbourClicked();
	void handleCapatibilitiesLoaded(QList<Capability> capabilities);
	void handleNeighbourMenuItemToggled(bool checked);
	void handleNeighbourMacEdit(bool toEditMode);
	void handleNeighbourAdvancedViewToggled(bool advanced);
	void callNeighbourMacEditDisable();

private:
	void setNoNotificationsVisible(bool visible);
	void setNoNeighboursVisible(bool visible);
	void createMenus();
	QSystemTrayIcon *trayIcon;
	void updateWindowPosition();
	void reinitConnection();
	void setNotificationsEnabled(bool enabled);
	void initCharts();
	void setDockMode(bool active);
	void enableAllTabsAndMenus();
	void recalculateSize();
	void updateTabsAndMenus(QList<Capability> capabilities);
	void setTabAndMenusEnabled(int tabIndex, bool enabled);
	void updateNeighbourDevices(QList<QString> devices);

	// Actions for context menu
	QAction *quitAction;
	QAction *settingsAction;
	QAction *notificationAction;
	QAction *statisticsAction;
	QAction *aboutAction;
	QAction *deviceAction;
	QAction *updaterAction;
	QAction *neighboursAction;
	QList<QAction *> tabActions;

	InterfaceWidget *interfaceWidget;
	UpdaterUserListsWidget *updaterUserList;
	UpdaterActivitiesWidget *updaterActivities;

	Ui::MainWindow *ui;
	int popupTime;
	QMenu *neighbourInterfaceMenu;
	QMenuBar *menuBar;
	Connector *connector;
	QTimer reconnectTimer;
	QTimer refreshTimer;
	bool systrayModeActive;
	bool dockModeActive;
	bool enableNeighbourChat;

signals:
	void translationNeeded();
};


#endif
