/*
 * Spectator - a desktop app which communicates with router Turris
 * Copyright (C) 2015  CZ.NIC, z. s. p. o. <http://www.nic.cz>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <QLayout>
#include <QFileDialog>
#include "consts.h"
#include "logging.h"
#include "settings.h"
#include "ui_settings.h"

SettingsDialog::SettingsDialog(QWidget *parent) :
	QDialog(parent),
	ui(new Ui::SettingsDialog)
{
	this->ui->setupUi(this);
	this->setFixedSize(this->size());
	this->ui->languageComboBox->addItem(tr("Czech"), Languages::czech);
	this->ui->languageComboBox->addItem(tr("English"), Languages::english);
}

SettingsDialog::~SettingsDialog()
{
	delete this->ui;
}

void SettingsDialog::displayTokenPathBrowseDialog()
{
	QString dirname = QFileInfo(this->ui->tokenPathEdit->text()).absoluteDir().absolutePath();
	QString res = QFileDialog::getOpenFileName(this, tr("token file path"), dirname, tr("Tokens (*.pem);;All files (*.*)"));
	if (!res.isNull()) {
		this->ui->tokenPathEdit->setText(res);
		this->ui->tokenPathEdit->setToolTip(res);
	}
}

void SettingsDialog::loadSettings(const SpectatorSettings &settings)
{
	this->ui->popupTimeSpinBox->setValue(settings.popupTime);
	this->ui->serverEdit->setText(settings.server);
	this->ui->refreshIntervalSpinBox->setValue(settings.refreshInterval);
	this->ui->portSpinBox->setValue(settings.port);
	int language_idx = this->ui->languageComboBox->findData(settings.language);
	if (language_idx < 0) {  // laguage not found -> use default
		language_idx = this->ui->languageComboBox->findData(DefaultSettings::language);
	}
	this->ui->languageComboBox->setCurrentIndex(language_idx);
	this->ui->tokenPathEdit->setText(settings.tokenPath);
	this->ui->tokenPathEdit->setToolTip(settings.tokenPath);
	this->ui->systrayModeCheckBox->setChecked(settings.systrayMode);
	this->ui->dockCheckBox->setChecked(settings.dock);
}

void SettingsDialog::updateSettings(SpectatorSettings &settings)
{
	settings.popupTime = this->ui->popupTimeSpinBox->value();
	settings.server = this->ui->serverEdit->text();
	settings.refreshInterval = this->ui->refreshIntervalSpinBox->value();
	settings.language = this->ui->languageComboBox->currentData().toString();
	settings.port = this->ui->portSpinBox->value();
	settings.tokenPath = this->ui->tokenPathEdit->text();
	settings.systrayMode = this->ui->systrayModeCheckBox->isChecked();
	settings.dock = this->ui->dockCheckBox->isChecked();
}

bool SettingsDialog::getDock()
{
	return this->ui->dockCheckBox->isChecked();
}

bool SettingsDialog::getSystrayMode()
{
	return this->ui->systrayModeCheckBox->isChecked();
}

SpectatorSettings::SpectatorSettings()
	: QSettings()
{
	this->beginGroup("Connection");
	// server is mandatory
	this->configPresent = !this->value("server").isNull();
	this->endGroup();
	if (this->configPresent)
		log_settings.debug("Previous configuration found.");
	else
		log_settings.debug("No previous configuration found. Using default settings.");
}

void SpectatorSettings::load()
{
	log_settings.debug("Loading settings.");

	this->beginGroup("Connection");
	this->server = this->value("server", DefaultSettings::server).toString();
	log_settings.debug() << "Loading server:" << this->server;
	this->port = this->value("port", DefaultSettings::port).toInt();
	log_settings.debug() << "Loading port:" << this->port;
	this->tokenPath = this->value("token_path", DefaultSettings::tokenPath).toString();
	log_settings.debug() << "Loading token_path:" << this->tokenPath;
	this->refreshInterval = this->value("refresh_interval", DefaultSettings::refreshInterval).toInt();
	log_settings.debug() << "Loading refresh_interval:" << this->refreshInterval;
	this->endGroup();

	this->beginGroup("GUI");
	this->popupTime = this->value("popup_time", DefaultSettings::popupTime).toInt();
	log_settings.debug() << "Loading popup_time:" << this->popupTime;
	this->mainTabIndex = this->value("main_tab_index", DefaultSettings::mainTabIndex).toInt();
	log_settings.debug() << "Loading main_tab_index:" << this->mainTabIndex;
	this->language = this->value("language", DefaultSettings::language).toString();
	log_settings.debug() << "Loading language:" << this->language;
	this->prefferedInterface = this->value("preffered_interface", DefaultSettings::preferedInterface).toString();
	log_settings.debug() << "Loading preffered_interface:" << this->prefferedInterface;
	this->systrayMode = this->value("systray_mode", DefaultSettings::systrayMode).toBool();
	log_settings.debug() << "Loading systray_mode:" << this->systrayMode;
	this->dock = this->value("dock", DefaultSettings::dock).toBool();
	log_settings.debug() << "Loading dock:" << this->dock;
	this->backupPath = this->value("backup_path", DefaultSettings::backupPath).toString();
	log_settings.debug() << "Loading backup_path:" << this->backupPath;
	this->geometry = this->value("geometry", QByteArray()).toByteArray();
	log_settings.debug() << "Loading geometry:" << this->geometry.toHex();
	this->endGroup();

	this->beginGroup("neighbours");
	this->inactiveLimit = this->value("neighbours_inactive_limit", DefaultSettings::inactiveLimit).toInt();
	log_settings.debug() << "Loading neighbours_inactive_limit:" << this->inactiveLimit;
	this->activeNeighbourDevices = this->value("neighbours_interfaces_filter", DefaultSettings::activeNeighbourDevices).toStringList();
	log_settings.debug() << "Loading neighbours_interfaces_filter:" << this->activeNeighbourDevices;
	QHash<QString, QVariant> macToName= this->value("neighbours_mac_to_name").toHash();
	this->neighboursMacToName.clear();
	foreach(QString macAddress, macToName.keys()) {
		this->neighboursMacToName.insert(macAddress, macToName.value(macAddress).toString());
	}
	log_settings.debug() << "Loading neighbours_mac_to_name:" << this->neighboursMacToName;
	this->neighboursAdvancedView = this->value("neighbours_advanced_view", DefaultSettings::neighboursAdvancedView).toBool();
	log_settings.debug() << "Loading neighbours_advanced_view:" << this->neighboursAdvancedView;
	this->endGroup();
}

void SpectatorSettings::store()
{
	log_settings.debug("Storing settings.");

	this->beginGroup("Connection");
	this->setValue("server", this->server);
	log_settings.debug() << "Storing server:" << this->server;
	this->setValue("port", this->port);
	log_settings.debug() << "Storing port:" << this->port;
	this->setValue("token_path", this->tokenPath);
	log_settings.debug() << "Storing token_path:" << this->tokenPath;
	this->setValue("refresh_interval", this->refreshInterval);
	log_settings.debug() << "Storing refresh_interval:" << this->refreshInterval;
	this->endGroup();

	this->beginGroup("GUI");
	this->setValue("popup_time", this->popupTime);
	log_settings.debug() << "Storing popup_time:" << this->popupTime;
	this->setValue("main_tab_index", this->mainTabIndex);
	log_settings.debug() << "Storing main_tab_index:" << this->mainTabIndex;
	this->setValue("language", this->language);
	log_settings.debug() << "Storing language:" << this->language;
	this->setValue("preffered_interface", this->prefferedInterface);
	log_settings.debug() << "Storing preffered_interface:" << this->prefferedInterface;
	this->setValue("systray_mode", this->systrayMode);
	log_settings.debug() << "Storing systray_mode:" << this->systrayMode;
	this->setValue("dock", this->dock);
	log_settings.debug() << "Storing dock:" << this->dock;
	this->setValue("backup_path", this->backupPath);
	log_settings.debug() << "Storing backup_path:" << this->backupPath;
	this->setValue("geometry", this->geometry);
	log_settings.debug() << "Storing geometry:" << this->geometry.toHex();
	this->endGroup();

	this->beginGroup("neighbours");
	this->setValue("neighbours_inactive_limit", this->inactiveLimit);
	log_settings.debug() << "Storing neighbours_inactive_limit" << this->inactiveLimit;
	this->setValue("neighbours_interfaces_filter", this->activeNeighbourDevices);
	log_settings.debug() << "Storing neighbours_interfaces_filter" << this->activeNeighbourDevices;

	QHash<QString, QVariant> macToName;
	foreach (QString macAddress, this->neighboursMacToName.keys()) {
		QString res = this->neighboursMacToName.value(macAddress);
		macToName.insert(macAddress, this->neighboursMacToName.value(macAddress));
	}
	this->setValue("neighbours_mac_to_name", macToName);
	log_settings.debug() << "Storing neighbours_mac_to_name" << this->neighboursMacToName;
	this->setValue("neighbours_advanced_view", this->neighboursAdvancedView);
	log_settings.debug() << "Storing neighbours_advanced_view" << this->neighboursAdvancedView;

	this->endGroup();
}
